import contextlib
import json
import queue
import re

from collections import OrderedDict
from datetime import datetime
from threading import Timer
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)
from urllib.parse import parse_qs, urlparse

from platypush.entities import (
    DimmerEntityManager,
    EnumSwitchEntityManager,
    Entity,
    LightEntityManager,
    SensorEntityManager,
    SwitchEntityManager,
)
from platypush.entities.batteries import Battery
from platypush.entities.devices import Device
from platypush.entities.dimmers import Dimmer
from platypush.entities.electricity import (
    EnergySensor,
    PowerSensor,
    VoltageSensor,
)
from platypush.entities.humidity import HumiditySensor
from platypush.entities.illuminance import IlluminanceSensor
from platypush.entities.lights import Light
from platypush.entities.sensors import BinarySensor, EnumSensor, NumericSensor
from platypush.entities.switches import EnumSwitch, Switch
from platypush.entities.temperature import TemperatureSensor
from platypush.message.event.zwave import (
    ZwaveEvent,
    ZwaveNodeAddedEvent,
    ZwaveNodeAsleepEvent,
    ZwaveNodeAwakeEvent,
    ZwaveNodeEvent,
    ZwaveNodeReadyEvent,
    ZwaveNodeRemovedEvent,
    ZwaveNodeRenamedEvent,
    ZwaveValueChangedEvent,
    ZwaveValueRemovedEvent,
)

from platypush.context import get_bus
from platypush.message.response import Response
from platypush.plugins import RunnablePlugin
from platypush.plugins.mqtt import MqttPlugin, action
from platypush.plugins.zwave._constants import command_class_by_name

from ._state import IdType, State


# pylint: disable=too-many-ancestors
class ZwaveMqttPlugin(
    MqttPlugin,
    RunnablePlugin,
    DimmerEntityManager,
    EnumSwitchEntityManager,
    LightEntityManager,
    SensorEntityManager,
    SwitchEntityManager,
):
    """
    This plugin allows you to manage a Z-Wave network over MQTT through
    `zwave-js-ui <https://github.com/zwave-js/zwave-js-ui>`_.

    Configuration required on the zwave-js-ui gateway:

        * Install the gateway following the instructions reported
          `here <https://zwave-js.github.io/zwave-js-ui/#/getting-started/quick-start>`_.

        * Configure the following settings from the zwave-js-ui web panel:

            * Zwave -> Serial Port: The path to your Z-Wave adapter.
            * Disabled MQTT Gateway: Set to false.
            * Configure the MQTT server that should be used by the service.
            * Mqtt -> Name: Unique name that identifies your gateway.
            * Mqtt -> Prefix: Prefix name for the topics published to MQTT.
            * Gateway -> Topic type: Set to "ValueID topics".
            * Gateway -> Payload type: Set to "JSON Time-Value".
            * Gateway -> Use nodes name instead of numeric nodeIDs: Set to false.
            * Gateway -> Send Zwave Events: Set to true.
            * Gateway -> Include Node Info: Set to true.

    """

    # These classes are ignored by the entity parsing logic
    _ignored_entity_classes = {
        'application_status',
        'association_command_configuration',
        'controller_replication',
        'crc16_encap',
        'firmware_update_md',
        'grouping_name',
        'ip_configuration',
        'manufacturer_proprietary',
        'manufacturer_specific',
        'multi_cmd',
        'multi_instance',
        'multi_instance_association',
        'no_operation',
        'node_naming',
        'non_interoperable',
        'proprietary',
        'scene_actuator_conf',
        'scene_controller_conf',
        'sensor_configuration',
        'version',
        'wake_up',
        'zwave_plus_info',
    }

    def __init__(
        self,
        name: str,
        host: str,
        port: int = 1883,
        topic_prefix: str = 'zwave',
        timeout: int = 10,
        tls_certfile: Optional[str] = None,
        tls_keyfile: Optional[str] = None,
        tls_version: Optional[str] = None,
        tls_ciphers: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ):
        """
        :param name: Gateway name, as configured from the zwavejs2mqtt web
            panel from Mqtt -> Name.
        :param host: MQTT broker host, as configured from the zwavejs2mqtt web
            panel from Mqtt -> Host
        :param port: MQTT broker listen port, as configured from the
            zwavejs2mqtt web panel from Mqtt -> Port (default: 1883).
        :param topic_prefix: MQTT topic prefix, as specified from the
            zwavejs2mqtt web panel from Mqtt -> Prefix (default: ``zwave``).
        :param timeout: If the command expects from a response, then this
            timeout value will be used (default: 10 seconds).
        :param tls_cafile: If the connection requires TLS/SSL, specify the
            certificate authority file (default: None)
        :param tls_certfile: If the connection requires TLS/SSL, specify the
            certificate file (default: None)
        :param tls_keyfile: If the connection requires TLS/SSL, specify the key
            file (default: None)
        :param tls_version: If the connection requires TLS/SSL, specify the
            minimum TLS supported version (default: None)
        :param tls_ciphers: If the connection requires TLS/SSL, specify the
            supported ciphers (default: None)
        :param username: If the connection requires user authentication,
            specify the username (default: None)
        :param password: If the connection requires user authentication,
            specify the password (default: None)
        """

        self.topic_prefix = topic_prefix
        self.base_topic = topic_prefix + '/{}/ZWAVE_GATEWAY-' + name
        self.events_topic = self.base_topic.format('_EVENTS')
        super().__init__(
            host=host,
            port=port,
            topics=[
                self.events_topic + '/node/' + topic
                for topic in [
                    'node_ready',
                    'node_sleep',
                    'node_value_updated',
                    'node_metadata_updated',
                    'node_wakeup',
                ]
            ],
            tls_certfile=tls_certfile,
            tls_keyfile=tls_keyfile,
            tls_version=tls_version,
            tls_ciphers=tls_ciphers,
            username=username,
            password=password,
            timeout=timeout,
            **kwargs,
        )

        self._state = State()

    def _api_topic(self, api: str) -> str:
        return self.base_topic.format('_CLIENTS') + f'/api/{api}'

    @staticmethod
    def _parse_response(response: Union[dict, Response]) -> dict:
        if isinstance(response, Response):
            assert not response.is_error(), response.errors[0]

        rs: dict = (
            response.output if isinstance(response, Response) else response
        )  # type: ignore

        assert rs.get('success') is True, rs.get('message', 'zwavejs2mqtt error')
        return rs

    def _api_request(self, api: str, *args: Any, **kwargs):
        if len(args) == 1 and isinstance(args[0], dict):
            args = args[0]  # type: ignore

        payload = json.dumps({'args': args})
        ret = self._parse_response(
            self.publish(
                topic=self._api_topic(api) + '/set',
                msg=payload,
                reply_topic=self._api_topic(api),
                **self._mqtt_args(**kwargs),
            )
        )

        assert not ret or ret.get('success') is True, ret.get('message')
        return ret.get('result')

    @staticmethod
    def _convert_timestamp(t: Optional[int]) -> Optional[datetime]:
        if t:
            return datetime.fromtimestamp(t / 1000)
        return None

    def _get_scene(
        self,
        scene_id: Optional[int] = None,
        scene_label: Optional[str] = None,
        **kwargs,
    ) -> dict:
        assert not (
            scene_id is None and scene_label is None
        ), 'No scene_id/scene_label specified'

        scene = self._state.scenes.get((scene_id, scene_label))
        if scene:
            return scene

        self._refresh_scenes(**kwargs)
        scene = self._state.scenes.get((scene_id, scene_label))
        assert scene, f'No such scene: scene_id={scene_id}, scene_label={scene_label}'
        return scene

    @classmethod
    def group_to_dict(cls, group: dict, node_id: Optional[int] = None) -> dict:
        return {
            'label': group['text'],
            'index': group['value'],
            'node_id': node_id,
            'group_id': f'{node_id}:{group["value"]}' if node_id is not None else None,
            'multichannel': group['multiChannel'],
            'max_associations': group['maxNodes'],
        }

    @classmethod
    def _get_type(cls, value: dict) -> str:
        if value.get('list'):
            if len(value.get('states', [])) > 1:
                return 'List'
            if value.get('min') is not None and value.get('max') is not None:
                return 'Decimal'
        if value['type'] == 'boolean':
            return 'Bool'
        if value['type'] == 'number':
            return 'Decimal'
        if value['type'] == 'string':
            return 'String'
        return value['type'].capitalize()

    @classmethod
    def value_to_dict(cls, value: dict) -> dict:
        is_read_only = (
            value['is_read_only']
            if 'is_read_only' in value
            else value.get('readable') and not value.get('writeable')
        )

        is_write_only = (
            value['is_write_only']
            if 'is_write_only' in value
            else not value.get('readable') and value.get('writeable')
        )

        return {
            'id': value['id'],
            'id_on_network': value['id'],
            'value_id': value['id'],
            'data': value.get('value'),
            'data_items': value['states'] if len(value.get('states', [])) > 1 else None,
            'label': value.get(
                'label', value.get('propertyName', value.get('property'))
            ),
            'property_id': value.get('property'),
            'help': value.get('description', value.get('help')),
            'node_id': value.get('nodeId'),
            'parent_id': value.get('nodeId'),
            'type': cls._get_type(value),
            'default': value.get('default'),
            'endpoint': value.get('endpoint'),
            'stateless': value.get('stateless'),
            'command_class': value.get('commandClass'),
            'command_class_name': value.get('commandClassName'),
            'units': value.get('unit'),
            'min': value.get('min'),
            'max': value.get('max'),
            'is_read_only': is_read_only,
            'is_write_only': is_write_only,
            'last_update': cls._convert_timestamp(value.get('lastUpdate')),
            **(
                {'property_key': value['propertyKey']} if 'propertyKey' in value else {}
            ),
            **(
                {'property_key_name': value['propertyKeyName']}
                if 'propertyKeyName' in value
                else {}
            ),
        }

    @staticmethod
    def _get_value_id(value: dict) -> str:
        return f'{value["nodeId"]}-{value["commandClass"]}-{value.get("endpoint", 0)}-{value["property"]}'

    @classmethod
    def scene_to_dict(cls, scene: dict) -> dict:
        return {
            'scene_id': scene.pop('sceneid', None),
            'values': {
                cls._get_value_id(value): {
                    'node_id': value['nodeId'],
                    'value_id': cls._get_value_id(value),
                    'id_on_network': cls._get_value_id(value),
                    'data': value.get('value'),
                }
                for value in scene.pop('values', [])
            },
            **scene,
        }

    @classmethod
    def node_to_dict(cls, node: dict) -> dict:
        capabilities = []
        if node.get('supportsBeaming'):
            capabilities += ['beaming']
        if node.get('supportsSecurity'):
            capabilities += ['security']
        if node.get('isRouting'):
            capabilities += ['routing']
        if node.get('zwavePlusVersion'):
            capabilities += ['zwave_plus']

        db_link = node.get('dbLink', node.get('db_link', ''))
        device_id = node.get('hexId')
        if not device_id:
            if db_link:
                # noinspection PyTypeChecker
                device_id = ':'.join(
                    parse_qs(urlparse(db_link).query)
                    .get('jumpTo', '')[0]
                    .split(':')[:-1]
                )
            else:
                device_id = str(node['id'])

        return {
            'node_id': node['id'],
            'device_id': device_id.replace('0x', ''),
            'name': node.get('name'),
            'capabilities': capabilities,
            'manufacturer_id': (
                f'0x{node["manufacturerId"]:04x}'
                if node.get('manufacturerId')
                else None
            ),
            'manufacturer_name': node.get('manufacturer'),
            'location': node.get('loc'),
            'status': node.get('status'),
            'is_available': node.get('available'),
            'is_awake': node.get('ready') and node.get('status') != 'Asleep',
            'is_beaming_device': node.get('supportsBeaming'),
            'is_controller': node.get('isControllerNode'),
            'is_failed': node.get('failed'),
            'is_frequent_listening_device': node.get('isFrequentListening'),
            'is_info_received': node.get('interviewStage') == 'Complete',
            'is_initialized': node.get('inited'),
            'is_listening_device': node.get('isListening'),
            'is_ready': node.get('ready'),
            'is_routing_device': node.get('isRouting'),
            'is_security_device': node.get('supportsSecurity'),
            'is_sleeping': node.get('ready') and node.get('status') == 'Asleep',
            'last_update': cls._convert_timestamp(node.get('lastActive')),
            'product_id': (
                f'0x{node["productId"]:04x}' if node.get('productId') else None
            ),
            'product_type': (
                f'0x{node["productType"]:04x}' if node.get('productType') else None
            ),
            'product_name': ' '.join(
                [node.get('productLabel', ''), node.get('productDescription', '')]
            ),
            'baud_rate': node.get('dataRate'),
            'max_baud_rate': node.get('maxBaudRate', node.get('dataRate')),
            'device_class': node.get('deviceClass'),
            'specific': node.get('deviceClass', {}).get('specific'),
            'firmware_version': node.get('firmwareVersion'),
            'keep_awake': node.get('keepAwake'),
            'db_link': node.get('dbLink'),
            'type': node.get('zwavePlusNodeType'),
            'role': node.get('zwavePlusRoleType'),
            'zwave_plus_version': node.get('zwavePlusVersion'),
            'zwave_plus_node_type': node.get('zwavePlusNodeType'),
            'zwave_plus_role_type': node.get('zwavePlusRoleType'),
            'neighbours': node.get('neighbors', []),
            'command_classes': {
                value['commandClassName']
                for value in node.get('values', {}).values()
                if value.get('commandClassName')
            },
            'groups': {
                group['value']: cls.group_to_dict(group, node_id=node['id'])
                for group in node.get('groups', [])
            },
            'values': {
                value['id']: cls.value_to_dict(value)
                for value in node.get('values', {}).values()
            },
        }

    def _refresh_groups(self, **kwargs):
        nodes = self._get_nodes(**kwargs)
        groups = {
            group['group_id']: {
                **group,
                'associations': [
                    assoc['nodeId']
                    for assoc in (
                        self._api_request('getAssociations', node_id, group['index'])
                        or []
                    )
                ],
            }
            for node_id, node in nodes.items()
            for group in node.get('groups', {}).values()
        }

        self._state.groups.add(*groups.values(), overwrite=True)

    def _get_groups(self, **kwargs) -> Dict[IdType, dict]:
        self._refresh_groups(**kwargs)
        return self._state.groups.by_id

    def _refresh_scenes(self, **kwargs):
        scenes = {
            scene.get('sceneid'): self.scene_to_dict(scene)
            for scene in (self._api_request('_getScenes', **kwargs) or [])
        }

        self._state.scenes.add(*scenes.values(), overwrite=True)

    def _get_scenes(self, **kwargs) -> Dict[IdType, dict]:
        self._refresh_scenes(**kwargs)
        return self._state.scenes.by_id

    def _refresh_nodes(self, **kwargs):
        nodes = {
            node['id']: {'id': node['id'], **self.node_to_dict(node)}
            for node in (self._api_request('getNodes', **kwargs) or [])
        }

        values = [
            value
            for node in nodes.values()
            for value in node.get('values', {}).values()
        ]

        # Process events for the newly received nodes
        for node_id, node in nodes.items():
            if node_id not in self._state.nodes:
                self._dispatch_event(ZwaveNodeAddedEvent, node=node, fetch_node=False)
            elif node['name'] != self._state.nodes.get('name'):
                self._dispatch_event(ZwaveNodeRenamedEvent, node=node, fetch_node=False)

        # Check if any previous node is now missing
        for node_id, node in self._state.nodes.by_id.items():
            if node_id not in nodes:
                self._dispatch_event(ZwaveNodeRemovedEvent, node=node, fetch_node=False)

        self._state.nodes.add(*nodes.values(), overwrite=True)
        self._state.values.add(*values, overwrite=True)
        self.publish_entities([self._to_current_value(v) for v in values])

    def _get_nodes(self, **kwargs) -> Dict[IdType, dict]:
        self._refresh_nodes(**kwargs)
        return self._state.nodes.by_id

    def _get_node(
        self,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        use_cache: bool = True,
        make_request: bool = True,
        **kwargs,
    ) -> dict:
        assert node_id or node_name, 'Please provide either a node_id or node_name'
        if use_cache:
            node = self._state.nodes.get((node_id, node_name))
            if node:
                return node

        assert make_request, f'No such node: {node_id or node_name}'
        self._refresh_nodes(**kwargs)
        return self._get_node(
            node_id=node_id,
            node_name=node_name,
            use_cache=True,
            make_request=False,
            **kwargs,
        )

    def _get_value(
        self,
        value_id: Optional[str] = None,
        id_on_network: Optional[str] = None,
        value_label: Optional[str] = None,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        use_cache: bool = True,
        make_request: bool = True,
        **kwargs,
    ) -> Dict[str, Any]:
        # Unlike python-openzwave, value_id and id_on_network are the same on zwavejs2mqtt
        value_id = value_id or id_on_network
        assert (
            value_id or value_label
        ), 'Please provide either value_id, id_on_network or value_label'

        if use_cache:
            value = self._state.values.get((value_id, value_label))
            if value:
                if node_id or node_name:
                    node = self._state.nodes.get((node_id, node_name))
                    assert node, f'No such node: {node_id or node_name}'
                    assert (
                        value['node_id'] == node['id']
                    ), f'No such value on node {node_id or node_name}: {value_id or value_label}'

                return value

        assert make_request, f'No such value: {value_id or value_label}'
        self._refresh_nodes(**kwargs)
        return self._get_value(
            value_id=value_id,
            id_on_network=id_on_network,
            value_label=value_label,
            node_id=node_id,
            node_name=node_name,
            use_cache=True,
            make_request=False,
            **kwargs,
        )

    @staticmethod
    def _is_target_value(value: Mapping):
        return value.get('property_id') in {'targetValue', 'targetColor'}

    @staticmethod
    def _is_current_value(value: Mapping):
        return value.get('property_id') in {'currentValue', 'currentColor'}

    def _to_current_value(self, value: dict) -> dict:
        return self._get_associated_value(value, 'current')

    def _to_target_value(self, value: dict) -> dict:
        return self._get_associated_value(value, 'target')

    def _get_associated_value(self, value: dict, target: str) -> dict:
        check_func = (
            self._is_target_value if target == 'current' else self._is_current_value
        )

        if not check_func(value):
            return value

        replace_args = (
            ('target', 'current') if target == 'current' else ('current', 'target')
        )

        associated_value_id = '-'.join(
            value['value_id'].split('-')[:-1]
            + [value['value_id'].split('-')[-1].replace(*replace_args)]
        )

        associated_value = self._state.values.get(associated_value_id)
        if associated_value:
            value = associated_value

        return value.copy()

    @staticmethod
    def _matches_classes(value: Mapping, *names: str):
        classes = {command_class_by_name[name] for name in names}
        return value.get('command_class', '') in classes if value else False

    @classmethod
    def _is_switch(cls, value: Mapping):
        return value.get('type') == 'Bool' and not value.get('is_read_only')

    @classmethod
    def _is_dimmer(cls, value: Mapping):
        return value.get('type') == 'Decimal' and not value.get('is_read_only')

    @classmethod
    def _is_enum_switch(cls, value: Mapping):
        return value.get('type') == 'List' and not value.get('is_read_only')

    @classmethod
    def _is_light(cls, value: Mapping):
        return (
            cls._matches_classes(value, 'color')
            and not {'red', 'green', 'blue'}.difference(
                set(value.get('value', {}).keys())
            )
            and not value.get('is_read_only')
        )

    @classmethod
    def _is_configuration_value(cls, value: Mapping):
        return cls._matches_classes(value, 'configuration')

    @classmethod
    def _get_sensor_args(
        cls, value: Mapping
    ) -> Tuple[Optional[Type], Optional[Mapping]]:
        sensor_type, args = None, None

        if value.get('is_read_only'):
            if value.get('type') == 'Bool':
                sensor_type, args = (
                    BinarySensor,
                    {
                        'value': value.get('data', False),
                    },
                )
            elif value.get('type') == 'List':
                sensor_type, args = (
                    EnumSensor,
                    {
                        'value': value.get('data'),
                        'values': {
                            i['value']: i['text'] for i in value.get('data_items', [])
                        },
                    },
                )
            elif value.get('type') == 'Decimal':
                sensor_type = NumericSensor
                if re.search(r'\s*power$', value['property_id'], re.IGNORECASE):
                    sensor_type = PowerSensor
                if re.search(r'\s*voltage$', value['property_id'], re.IGNORECASE):
                    sensor_type = VoltageSensor
                elif re.search(
                    r'\s*consumption$', value['property_id'], re.IGNORECASE
                ) or re.search(r'Wh$', (value.get('units') or '')):
                    sensor_type = EnergySensor
                elif re.search(r'\s*temperature$', value['property_id'], re.IGNORECASE):
                    sensor_type = TemperatureSensor
                elif re.search(
                    r'\s*(humidity|moisture)$', value['property_id'], re.IGNORECASE
                ):
                    sensor_type = HumiditySensor
                elif re.search(
                    r'\s*(illuminance|luminosity)$', value['property_id'], re.IGNORECASE
                ):
                    sensor_type = IlluminanceSensor

                args = {
                    'value': value.get('data'),
                    'min': value.get('min'),
                    'max': value.get('max'),
                    'unit': value.get('units'),
                }

        return sensor_type, args

    @classmethod
    def _is_battery(cls, value: Mapping):
        return (
            cls._matches_classes(value, 'battery')
            and value.get('is_read_only')
            and value.get('type') == 'Decimal'
        )

    def _to_entity_args(self, value: dict) -> dict:
        value = self._to_current_value(value)
        args = {
            'id': value['id'],
            'name': value.get('label'),
            'description': value.get('help'),
            'is_read_only': value.get('is_read_only'),
            'is_write_only': value.get('is_write_only'),
            'is_configuration': self._is_configuration_value(value),
        }

        if value.get('last_update'):
            args['updated_at'] = value['last_update']
        return args

    def transform_entities(self, entities: Iterable[dict]) -> Collection[Entity]:
        transformed_entities = []

        for value in entities:
            if not value or self._matches_classes(value, *self._ignored_entity_classes):
                continue

            value = value.copy()
            current_value = target_value = None
            if self._is_current_value(value):
                current_value = value
                target_value = self._to_target_value(value)
            elif self._is_target_value(value):
                current_value = self._to_current_value(value)
                target_value = value

            if current_value and target_value:
                value = self._merge_current_and_target_values(
                    [current_value, target_value]
                )[0]

            entity_type = None
            entity_args = self._to_entity_args(value)
            sensor_type, sensor_args = self._get_sensor_args(value)

            if self._is_light(value):
                entity_type = Light
                color = value['value']
                entity_args['red'] = color['red']
                entity_args['green'] = color['green']
                entity_args['blue'] = color['blue']
            elif self._is_dimmer(value):
                entity_type = Dimmer
                entity_args['value'] = value['data']
                entity_args['min'] = value['min']
                entity_args['max'] = value['max']
            elif self._is_enum_switch(value):
                entity_type = EnumSwitch
                entity_args['value'] = value['data']
                entity_args['values'] = {
                    i['value']: i['text'] for i in value.get('data_items', [])
                }
            elif self._is_battery(value):
                entity_type = Battery
                entity_args['value'] = value['data']
                entity_args['min'] = value['min']
                entity_args['max'] = value['max']
                entity_args['unit'] = value.get('units', '%')
            elif self._is_switch(value):
                entity_type = Switch
                entity_args['state'] = value['data']
            elif sensor_args:
                entity_type = sensor_type
                entity_args.update(sensor_args)

            if entity_type:
                transformed_entities.append(entity_type(**entity_args))

        self._process_parent_entities(entities, transformed_entities)
        return transformed_entities

    def _process_parent_entities(
        self, values: Iterable[Mapping], entities: List[Entity]
    ):
        parent_entities: Dict[int, Device] = {}
        entities_by_id: Dict[str, Entity] = {str(e.id): e for e in entities}

        for value in values:
            if not (value and value.get('id')):
                continue

            entity = entities_by_id.get(value['id'])
            if not entity:
                continue

            parent_id = value.get('parent_id')
            if parent_id is None:
                continue

            node = self._state.nodes.get(parent_id)
            if not node:
                continue

            node_id = node['node_id']
            if node_id not in parent_entities:
                parent = parent_entities[node_id] = Device(
                    id=node['device_id'],
                    name=node.get('name'),
                    external_url=self._build_external_url(node),
                    reachable=(
                        node.get('is_available', False) and node.get('is_ready', False)
                    ),
                )
                entities.append(parent)

            parent = parent_entities[node_id]
            entity.parent = parent
            entity.reachable = parent.reachable

    @staticmethod
    def _build_external_url(node: dict) -> Optional[str]:
        manufacturer_id = node.get('manufacturer_id')
        product_id = node.get('product_id')
        product_type = node.get('product_type')
        firmware_version = node.get('firmware_version', '0.0')
        if not (manufacturer_id and product_id and product_type):
            return None

        return (
            'https://devices.zwave-js.io/?jumpTo='
            f'{manufacturer_id}:{product_type}:{product_id}:{firmware_version}'
        )

    @classmethod
    def _merge_current_and_target_values(cls, values: Iterable[dict]) -> List[dict]:
        values_by_id = OrderedDict({v.get('id'): v for v in values})

        new_values = OrderedDict()
        for value in values:
            value = value.copy()
            value_id = value.get('id')
            if not value_id:
                continue

            associated_value = None
            associated_property_id = None
            current_property_id = None
            current_value = None
            value_id_prefix = '-'.join(value_id.split('-')[:-1])

            if cls._is_current_value(value):
                associated_property_id = value['property_id'].replace(
                    'current', 'target'
                )
                current_property_id = value['property_id']
                current_value = value
            elif cls._is_target_value(value):
                associated_property_id = value['property_id'].replace(
                    'target', 'current'
                )
                current_property_id = associated_property_id

            if associated_property_id:
                associated_value_id = f'{value_id_prefix}-{associated_property_id}'
                associated_value = values_by_id.pop(associated_value_id, None)

                if cls._is_target_value(value):
                    current_value = associated_value

            if current_value and associated_value and current_property_id:
                value = value.copy()
                value_id = f'{value_id_prefix}-{current_property_id}'
                value['data'] = current_value.get('data')
                value['id'] = value['value_id'] = value['id_on_network'] = value_id
                value['is_read_only'] = value['is_write_only'] = False
                value['label'] = 'Value'
                value['property_id'] = current_property_id
                value['last_update'] = (
                    max(
                        value.get('last_update') or 0,
                        associated_value.get('last_update') or 0,
                    )
                    or None
                )

            new_values[value_id] = value

        return list(new_values.values())

    def _filter_values(
        self,
        command_classes: Optional[Iterable[str]] = None,
        filter_callback: Optional[Callable[[dict], bool]] = None,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        nodes = (
            [self._get_node(node_name=node_name, use_cache=False, **kwargs)]
            if node_id or node_name
            else self._get_nodes(**kwargs).values()
        )

        classes: set = {
            command_class_by_name[command_name]
            for command_name in (command_classes or [])
        }

        values = {}

        for node in nodes:
            if not node:
                continue

            for value in node.get('values', {}).values():
                if (classes and value.get('command_class') not in classes) or (
                    filter_callback and not filter_callback(value)
                ):
                    continue

                value = self._to_current_value(value)
                values[value['id_on_network']] = value

        entity_values = self._merge_current_and_target_values(values.values())
        self.publish_entities(entity_values)  # type: ignore
        return values

    def _get_group(
        self,
        group_id: Optional[str] = None,
        group_index: Optional[int] = None,
        **kwargs,
    ) -> dict:
        assert not (
            group_id is None and group_index is None
        ), 'No group_id/group_index specified'
        group_id = group_id or str(group_index)
        group = self._state.groups.get(group_id)
        if group:
            return group

        groups = self._get_groups(**kwargs)
        assert group_id in groups, f'No such group_id: {group_id}'
        return groups[group_id]

    def on_mqtt_message(self):
        def handler(_, __, msg):
            if not msg.topic.startswith(self.events_topic):
                return

            topic = (
                msg.topic[(len(self.events_topic) + 1) :].split('/').pop()  # noqa: E203
            )

            data = msg.payload.decode()
            if not data:
                return

            with contextlib.suppress(ValueError, TypeError):
                data = json.loads(data)['data']

            try:
                if topic == 'node_value_updated':
                    self._dispatch_event(
                        ZwaveValueChangedEvent, node=data[0], value=data[1]
                    )
                elif topic == 'node_metadata_updated':
                    self._dispatch_event(ZwaveNodeEvent, node=data[0])
                elif topic == 'node_sleep':
                    self._dispatch_event(ZwaveNodeAsleepEvent, node=data[0])
                elif topic == 'node_wakeup':
                    self._dispatch_event(ZwaveNodeAwakeEvent, node=data[0])
                elif topic == 'node_ready':
                    self._dispatch_event(ZwaveNodeReadyEvent, node=data[0])
                elif topic == 'node_removed':
                    self._dispatch_event(ZwaveNodeRemovedEvent, node=data[0])
            except Exception as e:
                self.logger.exception(e)

        return handler

    def _dispatch_event(
        self,
        event_type: Type[ZwaveEvent],
        node: dict,
        value: Optional[dict] = None,
        fetch_node: bool = True,
        **kwargs,
    ):
        node_id = node.get('id')
        assert node_id is not None, 'No node ID specified'

        if fetch_node:
            node = kwargs['node'] = self._get_node(node_id)
        else:
            kwargs['node'] = node

        node_values = node.get('values', {})

        if node and value:
            # Infer the value_id structure if it's not provided on the event
            value_id = value.get('id')
            if value_id is None:
                value_id = f"{value['commandClass']}-{value.get('endpoint', 0)}-{value['property']}"
                if 'propertyKey' in value:
                    value_id += '-' + str(value['propertyKey'])

                # Prepend the node_id to value_id if it's not available in node['values']
                # (compatibility with more recent versions of ZwaveJS that don't provide
                # the value_id on the events)
                if value_id not in node_values:
                    value_id = f"{node_id}-{value_id}"

            if value_id not in node_values:
                self.logger.warning(
                    'value_id %s not found on node %s', value_id, node_id
                )
                return

            if 'newValue' in value:
                node_values[value_id]['data'] = value['newValue']

            value = kwargs['value'] = node_values[value_id]

        if issubclass(event_type, ZwaveNodeEvent):
            # If this node_id wasn't cached before, then it's a new node
            if node_id not in self._state.nodes:
                event_type = ZwaveNodeAddedEvent
            # If the name has changed, we have a rename event
            elif node['name'] != (self._state.nodes.get(node_id) or {}).get('name'):
                event_type = ZwaveNodeRenamedEvent

        if event_type == ZwaveNodeRemovedEvent:
            # If the node has been removed, remove it from the cache
            self._state.nodes.pop(node_id, None)
        else:
            # Otherwise, update the cached instance
            self._state.nodes.add(node)
            values = node.get('values', {}).values()
            self._state.values.add(*values)

        evt = event_type(**kwargs)
        get_bus().post(evt)

        if (
            value
            and issubclass(event_type, ZwaveValueChangedEvent)
            and event_type != ZwaveValueRemovedEvent
        ):
            self.publish_entities([value])

    @action
    def status(self, *_, **kwargs) -> Dict[str, Any]:
        """
        Get the current status of the Z-Wave values.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(**kwargs)

    @action
    def controller_status(self, **kwargs) -> Dict[str, Any]:
        """
        Get the status of the controller.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        msg_queue: queue.Queue = queue.Queue()
        topic = f'{self.topic_prefix}/driver/status'
        client = self._get_client(**kwargs)

        def on_message(_, __, msg):
            if msg.topic != topic:
                return
            msg_queue.put(json.loads(msg.payload))

        client.on_message = on_message
        client.connect()
        client.subscribe(topic)
        client.loop_start()

        try:
            status = msg_queue.get(
                block=True, timeout=kwargs.get('timeout', self.timeout)
            )
        except queue.Empty as e:
            raise TimeoutError('The request timed out') from e
        finally:
            client.loop_stop()

        return {
            'state': status,
        }

    @action
    def add_node(  # pylint: disable=arguments-differ
        self,
        name: str,
        location: str = '',
        do_security: bool = False,
        timeout: int = 30,
        **kwargs,
    ):
        """
        Start the inclusion process to add a node to the network.

        :param name: Name of the node to be added.
        :param location: Location of the node (default: empty).
        :param do_security: Whether to initialize the Network Key on the device if it supports the Security CC
        :param timeout: How long the inclusion process should last, in seconds (default: 30). Specify zero or null
            for no timeout.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self._api_request(
            'startInclusion',
            0,
            {
                'name': name,
                'location': location,
                'force_security': do_security,
            },
            **kwargs,
        )

        if timeout:
            Timer(timeout, lambda: self._api_request('stopInclusion', **kwargs)).start()

    @action
    def remove_node(self, timeout: int = 30, **kwargs):
        """
        Remove a node from the network (or, better, start the exclusion process).

        :param timeout: How long the exclusion process should last, in seconds (default: 30).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self._api_request('startExclusion', **kwargs)
        if timeout:
            Timer(timeout, lambda: self._api_request('stopExclusion', **kwargs)).start()

    @action
    def remove_failed_node(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ):
        """
        Remove a failed node from the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name).get('node_id')

        assert node_id is not None, f'No such node_id: {node_id}'
        self._api_request('removeFailedNode', node_id, **kwargs)

    @action
    def replace_failed_node(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ):
        """
        Replace a failed node on the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name).get('node_id')

        assert node_id is not None, f'No such node_id: {node_id}'
        self._api_request('replaceFailedNode', node_id, **kwargs)

    @action
    def request_network_update(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ):
        """
        Request a network update to a node.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name).get('node_id')

        assert node_id is not None, f'No such node_id: {node_id}'
        self._api_request('refreshInfo', node_id, **kwargs)

    @action
    def request_node_neighbour_update(self, *_, **kwargs):
        """
        Request a neighbours list update.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self._api_request('refreshNeighbors', **kwargs)

    @action
    def get_nodes(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Optional[Dict[IdType, dict]]:
        """
        Get the nodes associated to the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).

        :return: List of paired devices. Example output:

            .. code-block:: json

                {
                  "1": {
                    "node_id": 1,
                    "device_id": "0115-0400-0001",
                    "name": "Controller",
                    "capabilities": [
                      "beaming",
                      "routing"
                    ],
                    "manufacturer_id": "0x0115",
                    "manufacturer_name": "Z-Wave.Me",
                    "location": "Living Room",
                    "status": "Alive",
                    "is_available": true,
                    "is_awake": true,
                    "is_beaming_device": true,
                    "is_controller": true,
                    "is_failed": false,
                    "is_frequent_listening_device": false,
                    "is_info_received": true,
                    "is_initialized": true,
                    "is_listening_device": true,
                    "is_ready": true,
                    "is_routing_device": true,
                    "is_security_device": false,
                    "is_sleeping": false,
                    "last_update": "2021-04-05T19:38:07.202000",
                    "product_id": null,
                    "product_type": null,
                    "product_name": "UZB Z-Wave USB Stick",
                    "baud_rate": 100000,
                    "max_baud_rate": null,
                    "device_class": {
                      "basic": 2,
                      "generic": 2,
                      "specific": 1
                    },
                    "specific": 1,
                    "firmware_version": null,
                    "keep_awake": false,
                    "db_link": "https://devices.zwave-js.io/?jumpTo=0x0115:0x0400:0x0001:0.0",
                    "zwave_plus_version": null,
                    "zwave_plus_node_type": null,
                    "zwave_plus_role_type": null,
                    "neighbours": [
                      3,
                      4
                    ],
                    "groups": {},
                    "values": {}
                  },
                  "4": {
                    "node_id": 4,
                    "device_id": "010f-0602-1003",
                    "name": "TV Switch",
                    "capabilities": [
                      "beaming",
                      "routing",
                      "zwave_plus"
                    ],
                    "manufacturer_id": "0x010f",
                    "manufacturer_name": "FibaroGroup",
                    "location": "Living Room",
                    "status": "Alive",
                    "is_available": true,
                    "is_awake": true,
                    "is_beaming_device": true,
                    "is_controller": false,
                    "is_failed": false,
                    "is_frequent_listening_device": false,
                    "is_info_received": true,
                    "is_initialized": true,
                    "is_listening_device": true,
                    "is_ready": true,
                    "is_routing_device": true,
                    "is_security_device": false,
                    "is_sleeping": false,
                    "last_update": "2021-04-06T00:07:44.081000",
                    "product_id": null,
                    "product_type": null,
                    "product_name": "Metered Wall Plug Switch",
                    "baud_rate": 100000,
                    "max_baud_rate": null,
                    "device_class": {
                      "basic": 4,
                      "generic": 16,
                      "specific": 1
                    },
                    "specific": 1,
                    "firmware_version": "3.2",
                    "keep_awake": false,
                    "db_link": "https://devices.zwave-js.io/?jumpTo=0x010f:0x0602:0x1003:3.2",
                    "zwave_plus_version": 1,
                    "zwave_plus_node_type": 0,
                    "zwave_plus_role_type": 5,
                    "neighbours": [
                      1
                    ],
                    "groups": {
                      "1": {
                        "label": "Lifeline",
                        "index": 1,
                        "multichannel": true,
                        "max_associations": 1
                      },
                      "2": {
                        "label": "On/Off (Button)",
                        "index": 2,
                        "multichannel": true,
                        "max_associations": 10
                      },
                      "3": {
                        "label": "On/Off (Power)",
                        "index": 3,
                        "multichannel": true,
                        "max_associations": 10
                      }
                    },
                    "values": {
                      "4-37-0-currentValue": {
                        "id": "4-37-0-currentValue",
                        "id_on_network": "4-37-0-currentValue",
                        "value_id": "4-37-0-currentValue",
                        "data": true,
                        "data_items": null,
                        "label": "Current value",
                        "property_id": "currentValue",
                        "help": null,
                        "node_id": 4,
                        "parent_id": 4,
                        "type": "Bool",
                        "default": null,
                        "endpoint": 0,
                        "stateless": false,
                        "command_class": 37,
                        "command_class_name": "Binary Switch",
                        "units": null,
                        "min": null,
                        "max": null,
                        "is_read_only": true,
                        "is_write_only": false,
                        "last_update": "2021-04-05T19:38:07.587000"
                      }
                    }
                  }
                }

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        if node_id or node_name:
            return self._get_node(
                node_id=node_id, node_name=node_name, use_cache=False, **kwargs
            )

        return self._get_nodes(**kwargs)

    @action
    def set_node_name(
        self,
        new_name: str,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Rename a node on the network.

        :param new_name: New name for the node.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name, **kwargs).get('node_id')

        assert node_id, f'No such node: {node_id}'
        self._api_request('setNodeName', node_id, new_name, **kwargs)
        get_bus().post(
            ZwaveNodeRenamedEvent(
                node={
                    **self._get_node(node_id=node_id, **kwargs),
                    'name': new_name,
                }
            )
        )

    @action
    def set_node_location(
        self,
        location: str,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Set the location of a node.

        :param location: Node location.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name, **kwargs).get('node_id')

        assert node_id, f'No such node: {node_id}'
        self._api_request('setNodeLocation', node_id, location, **kwargs)
        get_bus().post(
            ZwaveNodeEvent(
                node={
                    **self._get_node(node_id=node_id, **kwargs),
                    'location': location,
                }
            )
        )

    @action
    def heal(
        self, *_, timeout: Optional[int] = 60, **kwargs
    ):  # pylint: disable=arguments-differ
        """
        Heal network by requesting nodes rediscover their neighbours.

        :param timeout: Duration of the healing process in seconds (default: 60). Set to zero or null for no timeout.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self._api_request('beginHealingNetwork', **kwargs)
        if timeout:
            Timer(
                timeout, lambda: self._api_request('stopHealingNetwork', **kwargs)
            ).start()

    @action
    def get_value(
        self,
        value_id: Optional[str] = None,
        id_on_network: Optional[str] = None,
        value_label: Optional[str] = None,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Get a value on the network.

        :param value_id: Select by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._get_value(
            value_id=value_id,
            value_label=value_label,
            id_on_network=id_on_network,
            node_id=node_id,
            node_name=node_name,
            use_cache=False,
            **kwargs,
        )

    @action
    def set_value(
        self,
        *args,
        data=None,
        value_id: Optional[str] = None,
        id_on_network: Optional[str] = None,
        value_label: Optional[str] = None,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Set a value.

        :param data: Data to set for the value.
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        # Compatibility layer with the .set_value format used by
        # the entities frontend
        if args:
            value_id = args[0]

        value = self._get_value(
            value_id=value_id,
            value_label=value_label,
            id_on_network=id_on_network,
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )

        # Convert to target value if the value is in current value format,
        # as that would usually be the writeable attribute
        value = self._to_target_value(value)
        self._api_request(
            'writeValue',
            {
                'nodeId': value['node_id'],
                'commandClass': value['command_class'],
                'endpoint': value.get('endpoint', 0),
                'property': value['property_id'],
                **(
                    {'propertyKey': value['property_key']}
                    if 'property_key' in value
                    else {}
                ),
            },
            data,
            **kwargs,
        )

    @action
    def set_lights(self, lights, **kwargs):  # pylint: disable=arguments-differ
        """
        Set the state for one or more Z-Wave lights.
        """
        lights = [lights] if isinstance(lights, str) else lights
        for light in lights:
            self.set_value(light, kwargs)

    @action
    def node_heal(  # pylint: disable=arguments-differ
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ):
        """
        Heal network node by requesting the node to rediscover their neighbours.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name, **kwargs).get('node_id')
        self._api_request('healNode', node_id, **kwargs)

    @action
    def node_update_neighbours(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ):
        """
        Ask a node to update its neighbours table
        (same as :meth:`platypush.plugins.zwave.mqtt.ZwaveMqttPlugin.node_heal`).

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self.node_heal(node_id=node_id, node_name=node_name, **kwargs)

    @action
    def node_network_update(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ):
        """
        Update the controller with network information
        (same as :meth:`platypush.plugins.zwave.mqtt.ZwaveMqttPlugin.node_heal`).

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self.node_heal(node_id=node_id, node_name=node_name, **kwargs)

    @action
    def node_refresh_info(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ):
        """
        Fetch up-to-date information about the node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name, **kwargs).get('node_id')
        self._api_request('refreshInfo', node_id, **kwargs)

    @action
    def get_dimmers(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the dimmers on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['switch_multilevel', 'switch_toggle_multilevel'],
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )

    @action
    def get_node_config(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the configuration values of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            [
                'configuration',
                'ip_configuration',
                'association_command_configuration',
                'sensor_configuration',
            ],
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )

    @action
    def get_battery_levels(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the battery levels of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['battery'], node_id=node_id, node_name=node_name, **kwargs
        )

    @action
    def get_power_levels(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the power levels of this node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['powerlevel'], node_id=node_id, node_name=node_name, **kwargs
        )

    @action
    def get_bulbs(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the bulbs/LEDs on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['color'],
            filter_callback=lambda value: not value['is_read_only'],
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )

    @action
    def get_switches(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the switches on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['switch_binary', 'switch_toggle_binary', 'switch_all'],
            filter_callback=lambda value: not value['is_read_only'],
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )

    @action
    def get_sensors(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the sensors on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['sensor_binary', 'sensor_multilevel', 'sensor_alarm', 'meter'],
            filter_callback=lambda value: not value['is_write_only'],
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )

    @action
    def get_doorlocks(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the doorlocks on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['door_lock'], node_id=node_id, node_name=node_name, **kwargs
        )

    @action
    def get_locks(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the locks on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['lock'], node_id=node_id, node_name=node_name, **kwargs
        )

    @action
    def get_usercodes(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the usercodes on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['user_code'], node_id=node_id, node_name=node_name, **kwargs
        )

    @action
    def get_thermostats(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the thermostats on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            [
                'thermostat_heating',
                'thermostat_mode',
                'thermostat_operating_state',
                'thermostat_setpoint',
                'thermostat_fan_mode',
                'thermostat_fan_state',
                'thermostat_setback',
            ],
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )

    @action
    def get_protections(
        self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the protection-compatible devices on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        return self._filter_values(
            ['protection'], node_id=node_id, node_name=node_name, **kwargs
        )

    @action
    def get_groups(self, **kwargs) -> Dict[IdType, dict]:
        """
        Get the groups on the network.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).

        :return: A list of the available groups. Example:

          .. code-block:: json

            {
              "2-2": {
                "label": "Motion",
                "multichannel": true,
                "max_associations": 10,
                "group_id": "2-2",
                "node_id": 2,
                "index": 2,
                "associations": [
                  3,
                  4
                ]
              },
              "5-3": {
                "label": "Tamper",
                "multichannel": true,
                "max_associations": 10,
                "group_id": "5-3",
                "node_id": 5,
                "index": 3,
                "associations": [
                  2,
                  3,
                  4
                ]
              }
            ]

        """
        return self._get_groups(**kwargs)

    @action
    def get_scenes(self, **kwargs) -> Dict[IdType, dict]:
        """
        Get the scenes configured on the network.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        :return: dict with the following format:

          .. code-block:: json

            {
              1: {
                "scene_id": 1,
                "label": "Scene 1",
                "values": {
                  "4-37-0-targetValue": {
                     "node_id": 4,
                     "value_id": "4-37-0-targetValue"
                     "data": true
                  },
                  "3-112-0-Temperature": {
                    "node_id": 3,
                    "value_id": "3-112-0-Temperature",
                    "data": true
                  }
                }
              }
            }

        """
        return self._get_scenes(**kwargs)

    @action
    def create_scene(self, label: str, **kwargs):
        """
        Create a new scene.

        :param label: Scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self._api_request('_createScene', label, **kwargs)

    @action
    def remove_scene(
        self,
        scene_id: Optional[int] = None,
        scene_label: Optional[str] = None,
        **kwargs,
    ):
        """
        Remove a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)
        self._api_request('_removeScene', scene['scene_id'])

    @action
    def activate_scene(
        self,
        scene_id: Optional[int] = None,
        scene_label: Optional[str] = None,
        **kwargs,
    ):
        """
        Activate a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)
        self._api_request('_activateScene', scene['scene_id'])

    @action
    def scene_add_value(
        self,
        data: Optional[Any] = None,
        value_id: Optional[str] = None,
        id_on_network: Optional[str] = None,
        value_label: Optional[str] = None,
        scene_id: Optional[int] = None,
        scene_label: Optional[str] = None,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Add a value to a scene.

        :param data: Data to set for the value (default: current value data).
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        value = self._get_value(
            value_id=value_id,
            value_label=value_label,
            id_on_network=id_on_network,
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)

        self._api_request(
            '_addSceneValue',
            scene['scene_id'],
            {
                'nodeId': value['node_id'],
                'commandClass': value['command_class'],
                'property': value['property_id'],
                'endpoint': value['endpoint'],
            },
            data,
            kwargs.get('timeout', self.timeout),
        )

    @action
    def scene_remove_value(
        self,
        value_id: Optional[str] = None,
        id_on_network: Optional[str] = None,
        value_label: Optional[str] = None,
        scene_id: Optional[int] = None,
        scene_label: Optional[str] = None,
        node_id: Optional[int] = None,
        node_name: Optional[str] = None,
        **kwargs,
    ):
        """
        Remove a value from a scene.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        value = self._get_value(
            value_id=value_id,
            value_label=value_label,
            id_on_network=id_on_network,
            node_id=node_id,
            node_name=node_name,
            **kwargs,
        )
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)
        self._api_request('_removeSceneValue', scene['scene_id'], value['value_id'])

    @action
    def get_scene_values(
        self,
        scene_id: Optional[int] = None,
        scene_label: Optional[str] = None,
        **kwargs,
    ) -> dict:
        """
        Get the values associated to a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)
        return scene.get('values', {})

    @action
    def add_node_to_group(  # pylint: disable=arguments-differ
        self,
        group_id: Optional[str] = None,
        node_id: Optional[int] = None,
        endpoint: Optional[int] = None,
        **kwargs,
    ):
        """
        Add a node to a group.

        :param group_id: Group ID.
        :param node_id: Node ID to be added.
        :param endpoint: Add a specific endpoint of the node to the group (default: add a node association).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        group = self._get_group(group_id, **kwargs)
        assoc = {'nodeId': node_id}
        if endpoint is not None:
            assoc['endpoint'] = endpoint

        self._api_request('addAssociations', group['node_id'], group['index'], [assoc])

    @action
    def remove_node_from_group(  # pylint: disable=arguments-differ
        self,
        group_id: Optional[str] = None,
        node_id: Optional[int] = None,
        endpoint: Optional[int] = None,
        **kwargs,
    ):
        """
        Remove a node from a group.

        :param group_id: Group ID.
        :param node_id: Node ID to be added.
        :param endpoint: Node endpoint to remove (default: remove node association).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        group = self._get_group(group_id, **kwargs)
        assoc = {'nodeId': node_id}
        if endpoint is not None:
            assoc['endpoint'] = endpoint

        self._api_request(
            'removeAssociations', group['node_id'], group['index'], [assoc]
        )

    @action
    def hard_reset(self, **_):
        """
        Perform a hard reset of the controller. It erases its network configuration settings.
        The controller becomes a primary controller ready to add devices to a new network.
        """
        self._api_request('hardReset')

    @action
    def on(self, device: str, *_, **kwargs):  # pylint: disable=arguments-differ
        """
        Turn on a switch on a device.

        :param device: ``id_on_network`` of the value to be switched on.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self.set_value(data=True, id_on_network=device, **kwargs)

    @action
    def off(self, device: str, *_, **kwargs):  # pylint: disable=arguments-differ
        """
        Turn off a switch on a device.

        :param device: ``id_on_network`` of the value to be switched off.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        self.set_value(data=False, id_on_network=device, **kwargs)

    @action
    def toggle(  # pylint: disable=arguments-differ
        self, device: str, *_, **kwargs
    ) -> dict:
        """
        Toggle a switch on a device.

        :param device: ``id_on_network`` of the value to be toggled.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish`
            (default: query the default configured device).
        """
        value = self._get_value(id_on_network=device, use_cache=False, **kwargs)
        value['data'] = not value['data']
        self.set_value(data=value['data'], id_on_network=device, **kwargs)
        node = self._state.nodes.get(value['node_id'])
        assert node, f'Node {value["node_id"]} not found'

        return {
            'name': (node['name'] + ' - ' + value.get('label', '[No Label]')),
            'on': value['data'],
            'id': value['value_id'],
        }

    def main(self):
        self.get_nodes()
        super().main()


# vim:sw=4:ts=4:et:



import base64
from datetime import datetime

from marshmallow import EXCLUDE, fields, pre_dump
from marshmallow.schema import Schema

from platypush.schemas import StrippedString


class PipedVideoSchema(Schema):
    """
    Class for video items returned by the Piped API.
    """

    class Meta:
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    item_type = fields.Constant(
        'video',
        metadata={
            'description': 'Item type',
            'example': 'video',
        },
    )

    url = fields.Url(
        required=True,
        metadata={
            'description': 'Video URL',
            'example': 'https://youtube.com/watch?v=1234567890',
        },
    )

    title = StrippedString(
        missing='[No Title]',
        metadata={
            'description': 'Video title',
            'example': 'My Video Title',
        },
    )

    image = fields.Url(
        attribute='thumbnail',
        metadata={
            'description': 'Image URL',
            'example': 'https://i.ytimg.com/vi/1234567890/hqdefault.jpg',
        },
    )

    description = StrippedString(
        attribute='shortDescription',
        metadata={
            'description': 'Video description',
            'example': 'My video description',
        },
    )

    duration = fields.Int(
        missing=0,
        metadata={
            'description': 'Video duration in seconds',
            'example': 120,
        },
    )

    channel = fields.String(
        attribute='uploaderName',
        metadata={
            'description': 'Channel name',
            'example': 'My Channel',
        },
    )

    channel_url = fields.Url(
        attribute='uploaderUrl',
        metadata={
            'description': 'Channel URL',
            'example': 'https://youtube.com/channel/1234567890',
        },
    )

    channel_image = fields.Url(
        attribute='uploaderAvatar',
        metadata={
            'description': 'Channel image URL',
            'example': 'https://i.ytimg.com/vi/1234567890/hqdefault.jpg',
        },
    )

    created_at = fields.DateTime(
        attribute='uploaded',
        metadata={
            'description': 'Video upload date',
            'example': '2020-01-01T00:00:00',
        },
    )

    @pre_dump
    def fill_urls(self, data: dict, **_):
        for attr in ('url', 'uploaderUrl'):
            if data.get(attr) and not data[attr].startswith('https://'):
                data[attr] = f'https://youtube.com{data[attr]}'

        return data

    @pre_dump
    def normalize_timestamps(self, data: dict, **_):
        if data.get('uploaded') and isinstance(data['uploaded'], int):
            data['uploaded'] = datetime.fromtimestamp(data["uploaded"] / 1000)

        return data


class PipedPlaylistSchema(Schema):
    """
    Class for playlist items returned by the Piped API.
    """

    class Meta:
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    item_type = fields.Constant(
        'playlist',
        metadata={
            'description': 'Item type',
            'example': 'playlist',
        },
    )

    id = fields.UUID(
        required=True,
        metadata={
            'description': 'Playlist ID',
            'example': '12345678-1234-1234-1234-1234567890ab',
        },
    )

    name = StrippedString(
        missing='[No Name]',
        metadata={
            'description': 'Playlist name',
            'example': 'My Playlist Name',
        },
    )

    image = fields.Url(
        attribute='thumbnail',
        metadata={
            'description': 'Image URL',
            'example': 'https://i.ytimg.com/vi/1234567890/hqdefault.jpg',
        },
    )

    description = StrippedString(
        attribute='shortDescription',
        metadata={
            'description': 'Video description',
            'example': 'My video description',
        },
    )

    videos = fields.Int(
        missing=0,
        metadata={
            'description': 'Number of videos in the playlist',
            'example': 10,
        },
    )

    channel = fields.String(
        attribute='uploaderName',
        metadata={
            'description': 'Channel name',
            'example': 'My Channel',
        },
    )

    channel_url = fields.Url(
        attribute='uploaderUrl',
        metadata={
            'description': 'Channel URL',
            'example': 'https://youtube.com/channel/1234567890',
        },
    )

    channel_image = fields.Url(
        attribute='uploaderAvatar',
        metadata={
            'description': 'Channel image URL',
            'example': 'https://i.ytimg.com/vi/1234567890/hqdefault.jpg',
        },
    )

    url = fields.Url(
        metadata={
            'description': 'Playlist URL',
            'example': 'https://youtube.com/playlist?list=1234567890',
        },
    )

    @pre_dump
    def fill_urls(self, data: dict, **_):
        for attr in ('url', 'uploaderUrl'):
            if data.get(attr) and not data[attr].startswith('https://'):
                data[attr] = f'https://youtube.com{data[attr]}'

        if not data.get('id') and data.get('url'):
            data['id'] = data['url'].split('=')[-1]

        return data

    @pre_dump
    def normalize_timestamps(self, data: dict, **_):
        if data.get('uploaded') and isinstance(data['uploaded'], int):
            data['uploaded'] = datetime.fromtimestamp(data["uploaded"] / 1000)

        return data


class PipedChannelSchema(Schema):
    """
    Class for channel items returned by the Piped API.
    """

    class Meta:
        """
        Exclude unknown fields.
        """

        unknown = EXCLUDE

    id = fields.String(
        required=True,
        metadata={
            'description': 'Channel ID',
            'example': '1234567890',
        },
    )

    item_type = fields.Constant(
        'channel',
        metadata={
            'description': 'Item type',
            'example': 'channel',
        },
    )

    url = fields.String(
        required=True,
        metadata={
            'description': 'Channel URL',
            'example': 'https://youtube.com/channel/1234567890',
        },
    )

    name = StrippedString(
        missing='[No Name]',
        metadata={
            'description': 'Channel name',
            'example': 'My Channel Name',
        },
    )

    description = StrippedString(
        metadata={
            'description': 'Channel description',
            'example': 'My channel description',
        },
    )

    image = fields.Url(
        attribute='avatar',
        metadata={
            'description': 'Channel image URL',
            'example': 'https://i.ytimg.com/vi/1234567890/hqdefault.jpg',
        },
    )

    banner = fields.Url(
        attribute='bannerUrl',
        metadata={
            'description': 'Channel banner URL',
            'example': 'https://i.ytimg.com/vi/1234567890/hqdefault.jpg',
        },
    )

    subscribers = fields.Int(
        attribute='subscriberCount',
        missing=0,
        metadata={
            'description': 'Number of subscribers',
            'example': 1000,
        },
    )

    next_page_token = fields.String(
        attribute='nextpage',
        metadata={
            'description': 'The token that should be passed to get the next page of results',
            'example': '1234567890',
        },
    )

    items = fields.Nested(PipedVideoSchema, attribute='relatedStreams', many=True)

    @pre_dump
    def normalize_id_and_url(self, data: dict, **_):
        if data.get('id'):
            if not data.get('url'):
                data['url'] = f'https://youtube.com/channel/{data["id"]}'
        elif data.get('url'):
            data['id'] = data['url'].split('/')[-1]
            data['url'] = f'https://youtube.com{data["url"]}'
        else:
            raise AssertionError('Channel ID or URL not found')

        return data

    @pre_dump
    def normalize_avatar(self, data: dict, **_):
        if data.get('avatarUrl'):
            data['avatar'] = data.pop('avatarUrl')

        return data

    @pre_dump
    def serialize_next_page_token(self, data: dict, **_):
        if data.get('nextpage'):
            data['nextpage'] = base64.b64encode(data['nextpage'].encode()).decode()

        return data



from contextlib import contextmanager
import gzip
import json
import logging
from collections import defaultdict
from time import time
from threading import RLock
from typing import Dict, Optional

from platypush.backend import Backend
from platypush.message.event import Event
from platypush.message.response import Response
from platypush.plugins import Plugin
from platypush.utils import (
    get_backend_class_by_name,
    get_backend_name_by_class,
    get_plugin_class_by_name,
    get_plugin_name_by_class,
)

logger = logging.getLogger(__name__)


class Cache:
    """
    A cache for the parsed integration metadata.

    Cache structure:

      .. code-block:: python

        {
            <integration_category>: {
                <integration_type>: {
                    'doc': <integration_docstring>,
                    'args': {
                        <arg_name>: {
                            'name': <arg_name>,
                            'type': <arg_type>,
                            'doc': <arg_docstring>,
                            'default': <arg_default_value>,
                            'required': <arg_required>,
                        },
                        ...
                    },
                    'actions': {
                        <action_name>: {
                            'name': <action_name>,
                            'doc': <action_docstring>,
                            'args': {
                                ...
                            },
                            'returns': {
                                'type': <return_type>,
                                'doc': <return_docstring>,
                            },
                        },
                        ...
                    },
                    'events': [
                        <event_type1>,
                        <event_type2>,
                        ...
                    ],
                },
                ...
            },
            ...
        }

    """

    cur_version = 1.1
    """
    Cache version, used to detect breaking changes in the cache logic that require a cache refresh.
    """

    def __init__(
        self,
        items: Optional[Dict[type, Dict[type, dict]]] = None,
        saved_at: Optional[float] = None,
        loaded_at: Optional[float] = None,
        version: float = cur_version,
    ):
        self.saved_at = saved_at
        self.loaded_at = loaded_at
        self._cache: Dict[type, Dict[type, dict]] = defaultdict(dict)
        self._lock = RLock()
        self.version = version
        self.has_changes = False

        if items:
            self._cache.update(items)
            self.loaded_at = time()

    @classmethod
    def load(cls, cache_file: str) -> 'Cache':
        """
        Loads the components cache from disk.

        :param cache_file: Cache file path.
        """
        with gzip.open(cache_file, 'rb') as f:
            data = f.read()

        return cls.from_dict(json.loads(data.decode()))

    def dump(self, cache_file: str):
        """
        Dumps the components cache to disk.

        :param cache_file: Cache file path.
        """
        from platypush.message import Message

        self.version = self.cur_version
        self.saved_at = time()
        compressed_cache = gzip.compress(
            json.dumps(
                {
                    'saved_at': self.saved_at,
                    'version': self.version,
                    'items': self.to_dict(),
                },
                cls=Message.Encoder,
                sort_keys=True,
            ).encode()
        )

        with open(cache_file, 'wb') as f:
            f.write(compressed_cache)

        self.has_changes = False

    @classmethod
    def from_dict(cls, data: dict) -> 'Cache':
        """
        Creates a cache from a JSON-serializable dictionary.
        """
        return cls(
            items={
                Backend: {
                    k: v
                    for k, v in {
                        get_backend_class_by_name(backend_type): backend_meta
                        for backend_type, backend_meta in data.get('items', {})
                        .get('backends', {})
                        .items()
                    }.items()
                    if k
                },
                Plugin: {
                    k: v
                    for k, v in {
                        get_plugin_class_by_name(plugin_type): plugin_meta
                        for plugin_type, plugin_meta in data.get('items', {})
                        .get('plugins', {})
                        .items()
                    }.items()
                    if k
                },
                Event: data.get('items', {}).get('events', {}),
                Response: data.get('items', {}).get('responses', {}),
            },
            loaded_at=time(),
            saved_at=data.get('saved_at'),
            version=data.get('version', cls.cur_version),
        )

    def to_dict(self) -> Dict[str, Dict[str, dict]]:
        """
        Converts the cache items to a JSON-serializable dictionary.
        """
        return {
            'backends': {
                k: v
                for k, v in {
                    get_backend_name_by_class(backend_type): backend_meta
                    for backend_type, backend_meta in self.backends.items()
                }.items()
                if k
            },
            'plugins': {
                k: v
                for k, v in {
                    get_plugin_name_by_class(plugin_type): plugin_meta
                    for plugin_type, plugin_meta in self.plugins.items()
                }.items()
                if k
            },
            'events': {
                (k if isinstance(k, str) else f'{k.__module__}.{k.__qualname__}'): v
                for k, v in self.events.items()
                if k
            },
            'responses': {
                (k if isinstance(k, str) else f'{k.__module__}.{k.__qualname__}'): v
                for k, v in self.responses.items()
                if k
            },
        }

    def get(self, category: type, obj_type: Optional[type] = None) -> Optional[dict]:
        """
        Retrieves an object from the cache.

        :param category: Category type.
        :param obj_type: Object type.
        :return: Object metadata.
        """
        collection = self._cache[category]
        if not obj_type:
            return collection
        return collection.get(obj_type)

    def set(self, category: type, obj_type: type, value: dict):
        """
        Set an object on the cache.

        :param category: Category type.
        :param obj_type: Object type.
        :param value: Value to set.
        """
        self._cache[category][obj_type] = value
        self.has_changes = True

    @property
    def plugins(self) -> Dict[type, dict]:
        """Plugins metadata."""
        return self._cache[Plugin]

    @property
    def backends(self) -> Dict[type, dict]:
        """Backends metadata."""
        return self._cache[Backend]

    @property
    def events(self) -> Dict[type, dict]:
        """Events metadata."""
        return self._cache[Event]

    @property
    def responses(self) -> Dict[type, dict]:
        """Responses metadata."""
        return self._cache[Response]

    @contextmanager
    def lock(self):
        """
        Context manager that acquires a lock on the cache.
        """
        with self._lock:
            yield



from typing import Type

import sounddevice as sd

from platypush.message.event.sound import (
    SoundRecordingPausedEvent,
    SoundRecordingResumedEvent,
    SoundRecordingStartedEvent,
    SoundRecordingStoppedEvent,
)

from .._converters import RawInputAudioConverter
from .._model import AudioState, StreamType
from ._base import AudioThread


class AudioRecorder(AudioThread):
    """
    The ``AudioRecorder`` thread is responsible for recording audio from the
    input device, writing it to the converter process and dispatch the
    converted audio to the registered consumers.
    """

    def __init__(self, *args, output_format: str, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_format = output_format

    @property
    def direction(self) -> StreamType:
        return StreamType.INPUT

    def _audio_callback(self):
        # _ = frames
        # __ = time
        def callback(indata, outdata, _, __, status):
            if self.state != AudioState.RUNNING:
                return

            if status:
                self.logger.warning('Recording callback status: %s', status)

            if not self._converter:
                self.logger.warning(
                    'The ffmpeg converter process has already terminated'
                )
                self.notify_stop()
                raise sd.CallbackStop

            try:
                self._converter.write(indata.tobytes())
            except AssertionError as e:
                self.logger.warning('Audio converter callback error: %s', e)
                self.state = AudioState.STOPPED
                return

            if self.audio_pass_through:
                outdata[:] = indata

        return callback

    @property
    def _audio_converter_type(self) -> Type[RawInputAudioConverter]:
        return RawInputAudioConverter

    @property
    def _started_event_type(self) -> Type[SoundRecordingStartedEvent]:
        return SoundRecordingStartedEvent

    @property
    def _stopped_event_type(self) -> Type[SoundRecordingStoppedEvent]:
        return SoundRecordingStoppedEvent

    @property
    def _paused_event_type(self) -> Type[SoundRecordingPausedEvent]:
        return SoundRecordingPausedEvent

    @property
    def _resumed_event_type(self) -> Type[SoundRecordingResumedEvent]:
        return SoundRecordingResumedEvent

    @property
    def _converter_args(self) -> dict:
        return {
            'output_format': self.output_format,
            **super()._converter_args,
        }

    @property
    def _stream_args(self) -> dict:
        return {
            'callback': self._audio_callback(),
            **super()._stream_args,
        }


# vim:sw=4:ts=4:et:



import os
from typing import Tuple, Union, List, Optional

import psutil

from platypush.entities import Entity
from platypush.entities.devices import Device
from platypush.entities.managers import EntityManager
from platypush.entities.sensors import NumericSensor, PercentSensor
from platypush.entities.system import (
    Cpu,
    CpuInfo as CpuInfoModel,
    CpuStats as CpuStatsModel,
    CpuTimes as CpuTimesModel,
    Disk as DiskModel,
    MemoryStats as MemoryStatsModel,
    NetworkInterface as NetworkInterfaceModel,
    SwapStats as SwapStatsModel,
    SystemBattery,
    SystemFan,
    SystemTemperature,
)
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin
from platypush.schemas.system import (
    Battery,
    BatterySchema,
    ConnectionSchema,
    CpuFrequency,
    CpuFrequencySchema,
    CpuInfo,
    CpuInfoSchema,
    CpuStats,
    CpuStatsSchema,
    CpuTimes,
    CpuTimesSchema,
    Disk,
    DiskSchema,
    Fan,
    FanSchema,
    MemoryStats,
    MemoryStatsSchema,
    NetworkInterface,
    NetworkInterfaceSchema,
    Process,
    ProcessSchema,
    SwapStats,
    SwapStatsSchema,
    SystemInfoSchema,
    Temperature,
    TemperatureSchema,
    User,
    UserSchema,
)


# pylint: disable=too-many-ancestors
class SystemPlugin(SensorPlugin, EntityManager):
    """
    Plugin to get system info.
    """

    def __init__(self, *args, poll_interval: Optional[float] = 60, **kwargs):
        super().__init__(*args, poll_interval=poll_interval, **kwargs)
        self.__cpu_info: Optional[CpuInfo] = None

    @property
    def _cpu_info(self) -> CpuInfo:
        from cpuinfo import get_cpu_info

        if not self.__cpu_info:
            # The CPU information won't change while the process is running, so
            # it makes sense to cache it only once.
            self.__cpu_info = CpuInfoSchema().load(get_cpu_info())  # type: ignore

        return self.__cpu_info  # type: ignore

    @action
    def cpu_info(self):
        """
        Get CPU info.
        :return: .. schema:: system.CpuInfoSchema
        """
        return CpuInfoSchema().dump(self._cpu_info)

    @staticmethod
    def _load_cpu_times(times, many: bool) -> Union[CpuTimes, List[CpuTimes]]:
        return CpuTimesSchema().load(times, many=many)  # type: ignore

    @classmethod
    def _cpu_times_avg(cls, percent=True) -> CpuTimes:
        method = psutil.cpu_times_percent if percent else psutil.cpu_times
        times = method(percpu=False)
        return cls._load_cpu_times(times, many=False)  # type: ignore

    @classmethod
    def _cpu_times_per_cpu(cls, percent=True) -> List[CpuTimes]:
        method = psutil.cpu_times_percent if percent else psutil.cpu_times
        times = method(percpu=True)
        return cls._load_cpu_times(times, many=True)  # type: ignore

    @action
    def cpu_times(self, per_cpu=False, percent=True) -> Union[list, dict]:
        """
        Get the CPU times per status, either as absolute time or a percentage.

        :param per_cpu: Get per-CPU stats (default: False).
        :param percent: Get the stats in percentage (default: True).
        :return: If ``per_cpu=False``:

            .. schema:: system.CpuTimesSchema

        If ``per_cpu=True`` then a list will be returned, where each item
        identifies the CPU times of a core:

           .. schema:: system.CpuTimesSchema(many=True)

        """
        method = self._cpu_times_per_cpu if per_cpu else self._cpu_times_avg
        return CpuTimesSchema().dump(method(percent=percent), many=per_cpu)

    @action
    def cpu_percent(
        self, per_cpu: bool = False, interval: Optional[float] = None
    ) -> Union[float, List[float]]:
        """
        Get the CPU load percentage.

        :param per_cpu: Get per-CPU stats (default: False).
        :param interval: When *interval* is 0.0 or None compares system CPU times elapsed since last call or module
            import, returning immediately (non blocking). That means the first time this is called it will
            return a meaningless 0.0 value which you should ignore. In this case is recommended for accuracy that this
            function be called with at least 0.1 seconds between calls.
        :return: float if ``per_cpu=False``, ``list[float]`` otherwise.
        """
        percent = psutil.cpu_percent(percpu=per_cpu, interval=interval)

        if per_cpu:
            return list(percent)  # type: ignore
        return percent

    @staticmethod
    def _cpu_stats() -> CpuStats:
        return CpuStatsSchema().load(psutil.cpu_stats())  # type: ignore

    @action
    def cpu_stats(self) -> CpuStats:
        """
        Get CPU stats.

        :return: .. schema:: system.CpuStatsSchema
        """
        return CpuStatsSchema().dump(self._cpu_stats())  # type: ignore

    @staticmethod
    def _cpu_frequency_avg() -> CpuFrequency:
        return CpuFrequencySchema().load(psutil.cpu_freq(percpu=False))  # type: ignore

    @staticmethod
    def _cpu_frequency_per_cpu() -> List[CpuFrequency]:
        return CpuFrequencySchema().load(psutil.cpu_freq(percpu=True), many=True)  # type: ignore

    @action
    def cpu_frequency(
        self, per_cpu: bool = False
    ) -> Union[CpuFrequency, List[CpuFrequency]]:
        """
        Get the CPU frequency, in MHz.

        :param per_cpu: Get per-CPU stats (default: False).
        :return: If ``per_cpu=False``:

            .. schema:: system.CpuFrequencySchema

        If ``per_cpu=True`` then a list will be returned, where each item
        identifies the CPU times of a core:

           .. schema:: system.CpuFrequencySchema(many=True)

        """
        return self._cpu_frequency_per_cpu() if per_cpu else self._cpu_frequency_avg()

    @action
    def load_avg(self) -> Tuple[float, float, float]:
        """
        Get the average load as a vector that represents the load within the last 1, 5 and 15 minutes.
        """
        return psutil.getloadavg()

    @staticmethod
    def _mem_virtual() -> MemoryStats:
        return MemoryStatsSchema().load(psutil.virtual_memory())  # type: ignore

    @action
    def mem_virtual(self) -> dict:
        """
        Get the current virtual memory usage stats.

        :return: .. schema:: system.MemoryStatsSchema
        """
        return MemoryStatsSchema().dump(self._mem_virtual())  # type: ignore

    @staticmethod
    def _mem_swap() -> SwapStats:
        return SwapStatsSchema().load(psutil.swap_memory())  # type: ignore

    @action
    def mem_swap(self) -> dict:
        """
        Get the current swap memory usage stats.

        :return: .. schema:: system.SwapStatsSchema
        """
        return SwapStatsSchema().dump(self._mem_swap())  # type: ignore

    @staticmethod
    def _disk_info() -> List[Disk]:
        parts = {
            part.device: part._asdict()
            for part in psutil.disk_partitions()
            if part.fstype != 'squashfs'  # Exclude loopback mounts
        }

        basename_parts = {os.path.basename(part): part for part in parts}
        io_stats = {
            basename_parts[disk]: stats._asdict()
            for disk, stats in psutil.disk_io_counters(perdisk=True).items()
            if disk in basename_parts
        }

        usage = {
            disk: psutil.disk_usage(info['mountpoint'])._asdict()  # type: ignore
            for disk, info in parts.items()
        }

        return DiskSchema().load(  # type: ignore
            [
                {
                    **info,
                    **io_stats.get(part, {}),
                    **usage[part],
                }
                for part, info in parts.items()
            ],
            many=True,
        )

    @action
    def disk_info(self):
        """
        Get information about the detected disks and partitions.

        :return: .. schema:: system.DiskSchema(many=True)
        """
        return DiskSchema().dump(self._disk_info(), many=True)

    @staticmethod
    def _network_info() -> List[NetworkInterface]:
        addrs = psutil.net_if_addrs()
        stats = psutil.net_if_stats()

        return NetworkInterfaceSchema().load(  # type: ignore
            [
                {
                    'interface': interface,
                    'addresses': addrs.get(interface, []),
                    **(stats[interface]._asdict() if stats.get(interface) else {}),
                    **info._asdict(),
                }
                for interface, info in psutil.net_io_counters(pernic=True).items()
                if any(bool(val) for val in info._asdict().values())
            ],
            many=True,
        )

    @staticmethod
    def _network_info_avg() -> NetworkInterface:
        stats = psutil.net_io_counters(pernic=False)
        return NetworkInterfaceSchema().load(  # type: ignore
            {
                'interface': None,
                **stats._asdict(),
            }
        )

    @action
    def network_info(self, per_nic: bool = False):
        """
        Get the information and statistics for the network interfaces.

        :param per_nic: Return the stats grouped by interface (default: False).
        :return: If ``per_nic=False``:

            .. schema:: system.NetworkInterfaceSchema

        If ``per_nic=True`` then a list will be returned, where each item
        identifies the statistics per network interface:

           .. schema:: system.NetworkInterfaceSchema(many=True)
        """

        if per_nic:
            return NetworkInterfaceSchema().dump(self._network_info(), many=True)
        return NetworkInterfaceSchema().dump(self._network_info_avg())

    @action
    def network_connections(self, type: str = 'inet') -> List[dict]:
        """
        Get the list of active network connections.
        On MacOS this function requires root privileges.

        :param type: Connection type to filter (default: ``inet``). Supported
            types:

            +------------+----------------------------------------------------+
            | ``type``   | Description                                        |
            +------------+----------------------------------------------------+
            | ``inet``   | IPv4 and IPv6                                      |
            | ``inet4``  | IPv4                                               |
            | ``inet6``  | IPv6                                               |
            | ``tcp``    | TCP                                                |
            | ``tcp4``   | TCP over IPv4                                      |
            | ``tcp6``   | TCP over IPv6                                      |
            | ``udp``    | UDP                                                |
            | ``udp4``   | UDP over IPv4                                      |
            | ``udp6``   | UDP over IPv6                                      |
            | ``unix``   | UNIX socket (both UDP and TCP protocols)           |
            | ``all``    | Any families and protocols                         |
            +------------+----------------------------------------------------+

        :return: .. schema:: system.ConnectionSchema(many=True)
        """
        schema = ConnectionSchema()
        return schema.dump(
            schema.load(psutil.net_connections(kind=type), many=True),  # type: ignore
            many=True,
        )

    @staticmethod
    def _sensors_temperature() -> List[Temperature]:
        return TemperatureSchema().load(  # type: ignore
            [
                {
                    **sensor._asdict(),
                    'id': f'{kind}_{i + 1}',
                    'label': (f'{kind} #{i + 1}' if not sensor.label else sensor.label),
                }
                for kind, sensors in psutil.sensors_temperatures().items()
                for i, sensor in enumerate(sensors)
            ],
            many=True,
        )

    @action
    def sensors_temperature(self) -> List[dict]:
        """
        Get stats from the temperature sensors.

        :return: .. schema:: system.TemperatureSchema(many=True)
        """
        return TemperatureSchema().dump(self._sensors_temperature(), many=True)

    @staticmethod
    def _sensors_fan() -> List[Fan]:
        return FanSchema().load(  # type: ignore
            [
                {
                    **sensor._asdict(),
                    'id': f'{kind}_{i + 1}',
                    'label': (f'{kind} #{i + 1}' if not sensor.label else sensor.label),
                }
                for kind, sensors in psutil.sensors_fans().items()
                for i, sensor in enumerate(sensors)
            ],
            many=True,
        )

    @action
    def sensors_fan(self) -> List[dict]:
        """
        Get stats from the fan sensors.

        :return: .. schema:: system.FanSchema(many=True)
        """
        return FanSchema().dump(self._sensors_fan(), many=True)

    @staticmethod
    def _sensors_battery() -> Optional[Battery]:
        battery = psutil.sensors_battery()
        return BatterySchema().load(battery) if battery else None  # type: ignore

    @action
    def sensors_battery(self) -> Optional[dict]:
        """
        Get stats from the battery sensor.

        :return: .. schema:: system.BatterySchema
        """
        battery = self._sensors_battery()
        return BatterySchema().dump(battery) if battery else None  # type: ignore

    @staticmethod
    def _connected_users() -> List[User]:
        return UserSchema().load(psutil.users(), many=True)  # type: ignore

    @action
    def connected_users(self) -> List[dict]:
        """
        Get the list of connected users.

        :return: .. schema:: system.UserSchema
        """
        return UserSchema().dump(self._connected_users(), many=True)

    @classmethod
    def _processes(cls) -> List[Process]:
        """
        Get the list of running processes.

        :param filter: Filter the list by name.
        :return: List of :class:`platypush.message.response.system.ProcessResponse`.
        """
        return ProcessSchema().load(  # type: ignore
            filter(  # type: ignore
                lambda proc: proc is not None,
                [cls._get_process_if_exists(pid) for pid in psutil.pids()],
            ),
            many=True,
        )

    @action
    def processes(self) -> List[dict]:
        """
        Get the list of running processes.

        :return: .. schema:: system.ProcessSchema
        """
        return ProcessSchema().dump(self._processes(), many=True)

    @classmethod
    def _get_process_if_exists(cls, pid: int) -> Optional[psutil.Process]:
        try:
            return cls._get_process(pid)
        except psutil.NoSuchProcess:
            return None

    @staticmethod
    def _get_process(pid: int) -> psutil.Process:
        return psutil.Process(pid)

    @action
    def pid_exists(self, pid: int) -> bool:
        """
        :param pid: Process PID.
        :return: ``True`` if the process exists, ``False`` otherwise.
        """
        return psutil.pid_exists(pid)

    @action
    def suspend(self, pid: int):
        """
        Suspend a process.
        :param pid: Process PID.
        """
        self._get_process(pid).suspend()

    @action
    def resume(self, pid: int):
        """
        Resume a process.
        :param pid: Process PID.
        """
        self._get_process(pid).resume()

    @action
    def terminate(self, pid: int):
        """
        Terminate a process.
        :param pid: Process PID.
        """
        self._get_process(pid).terminate()

    @action
    def kill(self, pid: int):
        """
        Kill a process.
        :param pid: Process PID.
        """
        self._get_process(pid).kill()

    @action
    def wait(self, pid: int, timeout: Optional[int] = None):
        """
        Wait for a process to terminate.

        :param pid: Process PID.
        :param timeout: Timeout in seconds (default: ``None``).
        """
        self._get_process(pid).wait(timeout)

    @action
    def get_measurement(self, *_, **__):
        """
        :return: .. schema:: system.SystemInfoSchema
        """
        return SystemInfoSchema().dump(
            {
                'cpu': {
                    'frequency': self._cpu_frequency_avg(),
                    'info': self._cpu_info,
                    'load_avg': self.load_avg().output,  # type: ignore
                    'stats': self._cpu_stats(),
                    'times': self._cpu_times_avg(),
                    'percent': self.cpu_percent().output / 100.0,  # type: ignore
                },
                'memory': self._mem_virtual(),
                'swap': self._mem_swap(),
                'disks': self._disk_info(),
                'network': self._network_info(),
                'temperature': self._sensors_temperature(),
                'fans': self._sensors_fan(),
                'battery': self._sensors_battery(),
            }
        )

    def transform_entities(self, entities: dict) -> List[Entity]:
        cpu = entities['cpu'].copy()
        battery = entities['battery']

        return [
            Cpu(
                id='system:cpu',
                name='CPU',
                percent=cpu['percent'],
                children=[
                    CpuInfoModel(
                        id='system:cpu:info',
                        name='Info',
                        **cpu['info'],
                    ),
                    CpuStatsModel(
                        id='system:cpu:stats',
                        name='Statistics',
                        children=[
                            NumericSensor(
                                id=f'system:cpu:stats:{key}',
                                name=key,
                                value=value,
                            )
                            for key, value in cpu['stats'].items()
                        ],
                    ),
                    CpuTimesModel(
                        id='system:cpu:times',
                        name='Times',
                        children=[
                            PercentSensor(
                                id=f'system:cpu:times:{key}',
                                name=key,
                                value=time_percent,
                            )
                            for key, time_percent in cpu['times'].items()
                        ],
                    ),
                    Device(
                        id='system:cpu:load_avg',
                        name='Load Average',
                        children=[
                            NumericSensor(
                                id=f'system:cpu:load_avg:{mins}',
                                name=f'Last {mins} minute(s)',
                                value=round(val, 2),
                            )
                            for val, mins in zip(cpu['load_avg'], [1, 5, 15])
                        ],
                    ),
                    NumericSensor(
                        id='system:cpu:frequency',
                        name='Frequency',
                        value=round(cpu['frequency']['current'], 2),
                        min=cpu['frequency']['min'],
                        max=cpu['frequency']['max'],
                        unit='MHz',
                    ),
                ],
            ),
            MemoryStatsModel(
                id='system:memory',
                name='Memory',
                **entities['memory'],
            ),
            SwapStatsModel(
                id='system:swap',
                name='Swap',
                **entities['swap'],
            ),
            *[
                DiskModel(
                    id=f'system:disk:{disk["device"]}',
                    name=disk.pop('device'),
                    **disk,
                )
                for disk in entities['disks']
                if disk.get('device')
            ],
            *[
                NetworkInterfaceModel(
                    id=f'system:network_interface:{nic["interface"]}',
                    name=nic.pop('interface'),
                    reachable=nic.pop('is_up'),
                    **nic,
                )
                for nic in entities.get('network', [])
                if nic.get('interface')
            ],
            *[
                SystemTemperature(
                    id=f'system:temperature:{temp.pop("id")}',
                    name=temp.pop('label'),
                    unit='°C',
                    **temp,
                )
                for temp in entities.get('temperature', [])
                if temp.get('id') and temp.get('label')
            ],
            *[
                SystemFan(
                    id=f'system:fan:{fan.pop("id")}',
                    name=fan.pop('label'),
                    unit='rpm',
                    **fan,
                )
                for fan in entities.get('fans', [])
                if fan.get('id') and fan.get('label')
            ],
            *[
                SystemBattery(
                    id='system:battery',
                    name='Battery',
                    **battery,
                )
                if battery
                else ()
            ],
        ]


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import Dict, Type

from .argument import Argument


class Component(ABC):
    """
    Abstract interface for all the application components exposed through the
    `inspect` plugin.

    It includes integrations (plugins and backends) and messages (events and
    responses).
    """

    @staticmethod
    def _merge_params(params: Dict[str, Argument], new_params: Dict[str, Argument]):
        """
        Utility function to merge a new mapping of parameters into an existing one.
        """
        for param_name, param in new_params.items():
            # Set the parameter if it doesn't exist
            if param_name not in params:
                params[param_name] = param

            # Set the parameter documentation if it's not set
            if param.doc and not params[param_name].doc:
                params[param_name].doc = param.doc

            # If the new parameter has required=False,
            # then that should also be the value for the current ones
            if param.required is False:
                params[param_name].required = False

            # If the new parameter has a default value, and the current
            # one doesn't, then the default value should be set as the new one.
            if param.default is not None and params[param_name].default is None:
                params[param_name].default = param.default

    @classmethod
    @abstractmethod
    def by_name(cls, name: str) -> "Component":
        """
        :param name: Component type name.
        :return: A parsed component class given its name/type name.
        """

    @classmethod
    @abstractmethod
    def by_type(cls, type: Type) -> "Component":
        """
        :param type: Component type.
        :return: A parsed component class given its type.
        """

    @property
    @abstractmethod
    def cls(self) -> Type:
        """
        :return: The class of a component.
        """

    @property
    @abstractmethod
    def doc_url(self) -> str:
        """
        :return: The URL of the documentation of the component.
        """



import json
import logging
from typing import List, Tuple, Union

import numpy as np

from ...._utils import convert_nd_array
from ._base import SoundBase
from ._sound import Sound


class Mix(SoundBase):
    """
    This class models a set of mixed :class:`._sound.Sound` instances that can be played
    through an audio stream to an audio device
    """

    def __init__(self, *sounds, channels: int, dtype, **kwargs):
        super().__init__(**kwargs)
        self._sounds: List[Sound] = []
        self.logger = logging.getLogger(__name__)
        self.channels = channels
        self.dtype = np.dtype(dtype)

        for sound in sounds:
            self.add(sound)

    def __iter__(self):
        """
        Iterate over the object's attributes and return key-pair values.
        """
        for sound in self._sounds:
            yield dict(sound)

    def __str__(self):
        """
        Return a JSON string representation of the object.
        """
        return json.dumps(list(self))

    def add(self, *sounds: Union[Sound, dict]):
        """
        Add one or more sounds to the mix.
        """
        self._sounds += [Sound.build(sound) for sound in sounds]

    def remove(self, *sound_indices: int):
        """
        Remove one or more sounds from the mix.
        """
        assert self._sounds and all(
            0 <= sound_index < len(sound_indices) for sound_index in sound_indices
        ), f'Sound indices must be between 0 and {len(self._sounds) - 1}'

        for sound_index in sound_indices[::-1]:
            self._sounds.pop(sound_index)

    def get_wave(
        self,
        sample_rate: float,
        t_start: float = 0,
        t_end: float = 0,
        normalize_range: Tuple[float, float] = (-1.0, 1.0),
        on_clip: str = 'scale',
        **_,
    ):  # -> NDArray[np.number]:
        wave = None

        for sound in self._sounds:
            sound_wave = sound.get_wave(
                t_start=t_start, t_end=t_end, sample_rate=sample_rate
            )

            if wave is None:
                wave = sound_wave
            else:
                wave += sound_wave

        if wave is not None and len(wave):
            scale_factor = (normalize_range[1] - normalize_range[0]) / (
                wave.max() - wave.min()
            )

            if scale_factor < 1.0:  # Wave clipping
                if on_clip == 'scale':
                    wave = scale_factor * wave
                elif on_clip == 'clip':
                    wave[wave < normalize_range[0]] = normalize_range[0]
                    wave[wave > normalize_range[1]] = normalize_range[1]
                else:
                    raise RuntimeError(
                        'Supported values for "on_clip": ' + '"scale" or "clip"'
                    )

        assert wave is not None
        return convert_nd_array(self.gain * wave, dtype=self.dtype)

    def duration(self):
        """
        :returns: The duration of the mix in seconds as duration of its longest
            sample, or None if the mixed sample have no duration set
        """

        # If any sound has no duration specified, then the resulting mix will
        # have no duration as well.
        if any(sound.duration is None for sound in self._sounds):
            return None

        return max(((sound.duration or 0) + sound.delay for sound in self._sounds))


# vim:sw=4:ts=4:et:



from enum import IntEnum


class TransportEncryption(IntEnum):
    """
    Enum for mail transport encryption types.
    """

    NONE = 0
    STARTTLS = 1
    SSL = 2

    @classmethod
    def by_url_scheme(cls, scheme: str) -> 'TransportEncryption':
        """
        Get the transport encryption type from the specified URL scheme.
        """
        if scheme.endswith('+starttls'):
            return cls.STARTTLS
        if scheme.endswith('s'):
            return cls.SSL
        return cls.NONE


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from typing import Dict, Iterable, Optional

from platypush.plugins import Plugin, action


class MusicPlugin(Plugin, ABC):
    """
    Base class for music player plugins.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @action
    @abstractmethod
    def play(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def pause(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def stop(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def next(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def previous(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def set_volume(self, volume: float, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def volup(self, step: Optional[float] = None, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def voldown(self, step: Optional[float] = None, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def seek(self, position, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def add(self, resource, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def clear(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def status(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def current_track(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def get_playlists(self, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def get_playlist(self, playlist, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def add_to_playlist(self, playlist, resources, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def remove_from_playlist(self, playlist, resources, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def playlist_move(self, playlist, from_pos: int, to_pos: int, **kwargs):
        raise NotImplementedError()

    @action
    @abstractmethod
    def search(self, query, **kwargs):
        raise NotImplementedError()

    @action
    def get_images(self, resources: Iterable[str], **__) -> Dict[str, Optional[str]]:
        """
        Get the images for a list of URIs.

        .. note:: This is an optional action, and it may not be implemented by all plugins.
                  If the plugin doesn't implement this action, it will return an empty dictionary.

        :param uris: List of URIs.
        :return: Dictionary in the form ``{uri: image_url}``.
        """
        return {}


# vim:sw=4:ts=4:et:



from typing import Sequence
from platypush.message.event import Event, EventMatchResult


class FlicButtonEvent(Event):
    """
    Event triggered when a sequence of user short/long presses is detected on a
    `Flic button <https://flic.io>`_.
    """

    def __init__(self, btn_addr: str, sequence: Sequence[str], *_, **kwargs):
        """
        :param btn_addr: Physical address of the button that originated the
            event.
        :param sequence: Detected sequence, as a list of Flic button event
            types (either ``ShortPressEvent`` or ``LongPressEvent``).
        """
        super().__init__(btn_addr=btn_addr, sequence=sequence, **kwargs)

    def matches_condition(self, condition):
        """
        :param condition: Condition to be checked against, as a sequence of
            button presses (``ShortPressEvent`` and ``LongPressEvent``).
        """
        result = EventMatchResult(is_match=False)

        if (
            not isinstance(self, condition.type)
            or self.args['btn_addr'] != condition.args['btn_addr']
        ):
            return result

        cond_sequence = list(condition.args['sequence'])
        event_sequence = list(self.args['sequence'])

        while cond_sequence and event_sequence:
            cond_press = cond_sequence[0]
            event_press = event_sequence[0]

            if cond_press == event_press:
                cond_sequence.pop(0)
                result.score += 1

            event_sequence.pop(0)

        result.is_match = len(cond_sequence) == 0
        return result


# vim:sw=4:ts=4:et:



import logging
import time
from typing import Optional

from platypush.message.event.media import (
    MediaPlayEvent,
    MediaStopEvent,
    MediaPauseEvent,
    NewPlayingMediaEvent,
    MediaVolumeChangedEvent,
    MediaSeekEvent,
)

from ._utils import MediaCallback, convert_status, post_event

logger = logging.getLogger(__name__)


class MediaListener:
    """
    Listens for media status changes and posts events accordingly.
    """

    def __init__(self, name: str, cast, callback: Optional[MediaCallback] = None):
        self.name = name
        self.cast = cast
        self.status = convert_status(cast.media_controller.status)
        self.last_status_timestamp = time.time()
        self.callback = callback

    def new_media_status(self, status):
        status = convert_status(status)
        if status.get('url') and status.get('url') != self.status.get('url'):
            self._post_event(
                NewPlayingMediaEvent,
                title=status.get('title'),
            )

        state = status.get('state')
        if state != self.status.get('state'):
            if state == 'play':
                self._post_event(MediaPlayEvent)
            elif state == 'pause':
                self._post_event(MediaPauseEvent)
            elif state in ('stop', 'idle'):
                self._post_event(MediaStopEvent)

        if status.get('volume') != self.status.get('volume'):
            self._post_event(MediaVolumeChangedEvent, volume=status.get('volume'))

        if (
            status.get('position')
            and self.status.get('position')
            and abs(status['position'] - self.status['position'])
            > time.time() - self.last_status_timestamp + 5
        ):
            self._post_event(MediaSeekEvent, position=status.get('position'))

        self.status = status
        self.last_status_timestamp = time.time()

    def load_media_failed(self, item, error_code):
        logger.warning('Failed to load media %s: %d', item, error_code)

    def _post_event(self, evt_type, **evt):
        status = evt.get('status', {})
        resource = status.get('url')
        args = {
            'device': self.name,
            'plugin': 'media.chromecast',
            **evt,
        }

        if resource:
            args['resource'] = resource

        post_event(evt_type, callback=self.callback, chromecast=self.cast, **args)



from marshmallow import EXCLUDE, fields
from marshmallow.schema import Schema

from platypush.schemas import DateTime


class WeatherSchema(Schema):
    """
    Schema for weather data.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Weather schema metadata.
        """

        unknown = EXCLUDE

    time = DateTime(
        required=True,
        attribute='datetime',
        metadata={
            'description': 'Date and time of the weather data',
            'example': '2020-01-01T00:00:00+00:00',
        },
    )

    summary = fields.Function(
        lambda obj: obj.get('condition', {}).get('exact', 'Unknown'),
        metadata={
            'description': 'Summary of the weather condition',
            'example': 'Cloudy',
        },
    )

    image = fields.Function(
        lambda obj: obj.get('condition', {}).get('image'),
        metadata={
            'description': 'Image URL representing the weather condition',
            'example': 'https://www.buienradar.nl/resources/images/icons/weather/30x30/cc.png',
        },
    )

    precip_intensity = fields.Float(
        attribute='rainlasthour',
        metadata={
            'description': 'Amount of precipitation in the last hour in mm/h',
            'example': 0.0,
        },
    )

    precip_type = fields.Function(
        lambda obj: 'snow' if obj.get('snow') else 'rain',
        metadata={
            'description': 'Type of precipitation',
            'example': 'rain',
        },
    )

    temperature = fields.Float(
        metadata={
            'description': 'Temperature in Celsius',
            'example': 10.0,
        },
    )

    apparent_temperature = fields.Float(
        attribute='feeltemperature',
        metadata={
            'description': 'Apparent temperature in Celsius',
            'example': 9.0,
        },
    )

    humidity = fields.Float(
        metadata={
            'description': 'Humidity percentage, between 0 and 100',
            'example': 30,
        },
    )

    pressure = fields.Float(
        metadata={
            'description': 'Pressure in hPa',
            'example': 1000.0,
        },
    )

    rain_chance = fields.Float(
        attribute='rainchance',
        metadata={
            'description': 'Chance of rain in percentage, between 0 and 100',
            'example': 30,
        },
    )

    wind_speed = fields.Float(
        attribute='windspeed',
        metadata={
            'description': 'Wind speed in the configured unit of measure',
            'example': 10.0,
        },
    )

    wind_direction = fields.Float(
        attribute='windazimuth',
        metadata={
            'description': 'Wind direction in degrees',
            'example': 180,
        },
    )

    wind_gust = fields.Float(
        attribute='windgust',
        metadata={
            'description': 'Wind gust in the configured unit of measure',
            'example': 15.0,
        },
    )

    visibility = fields.Float(
        metadata={
            'description': 'Visibility in meters',
            'example': 2000.0,
        },
    )

    units = fields.Constant(
        'metric',
        metadata={
            'description': 'Unit of measure',
            'example': 'metric',
        },
    )


class PrecipitationForecastSchema(Schema):
    """
    Schema for precipitation forecast data.
    """

    class Meta:  # pylint: disable=too-few-public-methods
        """
        Precipitation forecast schema metadata.
        """

        unknown = EXCLUDE

    timeframe = fields.Integer(
        metadata={
            'description': 'Time frame in minutes',
            'example': 60,
        },
    )

    total = fields.Float(
        metadata={
            'description': 'Total precipitation in mm/h',
            'example': 0.5,
        },
    )

    average = fields.Float(
        metadata={
            'description': 'Average precipitation in mm/h',
            'example': 0.25,
        },
    )



from io import BytesIO
from threading import Event
from typing import Optional

import numpy as np
from pydub import AudioSegment

from platypush.common.assistant import AudioRecorder
from platypush.context import get_plugin
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.assistant import AssistantPlugin
from platypush.plugins.openai import OpenaiPlugin

from ._state import RecordingState


# pylint: disable=too-many-ancestors
class AssistantOpenaiPlugin(AssistantPlugin, RunnablePlugin):
    """
    A voice assistant based on the OpenAI API.

    It requires the :class:`platypush.plugins.openai.OpenaiPlugin` plugin to be
    configured with an OpenAI API key.

    Hotword detection
    -----------------

    This plugin doesn't have hotword detection, as OpenAI doesn't provide
    an API for that. Instead, the assistant can be started and stopped
    programmatically through the :meth:`.start_conversation` action.

    If you want to implement hotword detection, you can use a separate plugin
    such as
    :class:`platypush.plugins.assistant.picovoice.AssistantPicovoicePlugin`.

    The configuration in this case would be like:

        .. code-block:: yaml

            assistant.picovoice:
              access_key: YOUR_PICOVOICE_ACCESS_KEY

              # List of hotwords to listen for
              keywords:
                - alexa
                - computer
                - ok google

              # Disable speech-to-text and intent recognition, only use hotword
              # detection
              stt_enabled: false
              hotword_enabled: true

              conversation_start_sound: /sound/to/play/when/the/conversation/starts.mp3
              # speech_model_path: /mnt/hd/models/picovoice/cheetah/custom-en.pv
              # intent_model_path: /mnt/hd/models/picovoice/rhino/custom-en-x86.rhn

            openai:
              api_key: YOUR_OPENAI_API_KEY

              # Customize your assistant's context and knowledge base to your
              # liking
              context:
                - role: system
                  content: >
                    You are a 16th century noble lady who talks in
                    Shakespearean English to her peers.

            # Enable the assistant plugin
            assistant.openai:

            # Enable the text-to-speech plugin
            tts.openai:
              # Customize the voice model
              voice: nova

    Then you can call :meth:`.start_conversation` when the hotword is detected
    :class:`platypush.message.event.assistant.HotwordDetectedEvent` is
    triggered:

        .. code-block:: python

            from platypush import run, when
            from platypush.message.event.assistant import HotwordDetectedEvent

            @when(HotwordDetectedEvent)
            # You can also customize it by running a different assistant logic
            # depending on the hotword
            # @when(HotwordDetectedEvent, hotword='computer')
            def on_hotword_detected():
                run("assistant.openai.start_conversation")

    This configuration will:

        1. Start the hotword detection when the application starts.
        2. Start the OpenAI assistant when the hotword is detected.

    AI responses
    ------------

    By default (unless you set ``stop_conversation_on_speech_match`` to ``False``),
    the plugin will:

        1. Process the speech through the OpenAI API (the GPT model to be is
           configurable in the OpenAI plugin ``model`` configuration).

        2. Render the response through the configured ``tts_plugin`` (default:
           ``tts.openai``). If ``tts_plugin`` is not set, then the response will
           be returned as a string.

    Custom speech processing
    ------------------------

    You can create custom hooks on
    :class:`platypush.message.event.assistant.SpeechRecognizedEvent` with
    custom ``phrase`` strings or (regex) patterns. For example:

        .. code-block:: python

            from platypush import run, when
            from platypush.message.event.assistant import SpeechRecognizedEvent

            # Matches any phrase that contains either "play music" or "play the
            # music"
            @when(SpeechRecognizedEvent, phrase='play (the)? music')
            def play_music():
                run('music.mpd.play')

    If at least a custom hook with a non-empty ``phrase`` string is matched,
    then the default response will be disabled. If you still want the assistant
    to say something when the event is handled, you can call
    ``event.assistant.render_response`` on the hook:

        .. code-block:: python

            from datetime import datetime
            from textwrap import dedent
            from time import time

            from platypush import run, when
            from platypush.message.event.assistant import SpeechRecognizedEvent

            @when(SpeechRecognizedEvent, phrase='weather today')
            def weather_forecast(event: SpeechRecognizedEvent):
                limit = time() + 24 * 60 * 60  # 24 hours from now
                forecast = [
                    weather
                    for weather in run("weather.openweathermap.get_forecast")
                    if datetime.fromisoformat(weather["time"]).timestamp() < limit
                ]

                min_temp = round(
                    min(weather["temperature"] for weather in forecast)
                )
                max_temp = round(
                    max(weather["temperature"] for weather in forecast)
                )
                max_wind_gust = round(
                    (max(weather["wind_gust"] for weather in forecast)) * 3.6
                )
                summaries = [weather["summary"] for weather in forecast]
                most_common_summary = max(summaries, key=summaries.count)
                avg_cloud_cover = round(
                    sum(weather["cloud_cover"] for weather in forecast) / len(forecast)
                )

                event.assistant.render_response(
                    dedent(
                        f\"\"\"
                        The forecast for today is: {most_common_summary}, with
                        a minimum of {min_temp} and a maximum of {max_temp}
                        degrees, wind gust of {max_wind_gust} km/h, and an
                        average cloud cover of {avg_cloud_cover}%.
                        \"\"\"
                    )
                )

    Conversation follow-up
    ----------------------

    A conversation will have a follow-up (i.e. the assistant will listen for a
    phrase after rendering a response) if the response is not empty and ends
    with a question mark. If you want to force a follow-up even if the response
    doesn't end with a question mark, you can call :meth:`.start_conversation`
    programmatically from your hooks.
    """

    def __init__(
        self,
        model: str = "whisper-1",
        tts_plugin: Optional[str] = "tts.openai",
        min_silence_secs: float = 1.0,
        silence_threshold: int = -22,
        sample_rate: int = 16000,
        frame_size: int = 16384,
        channels: int = 1,
        conversation_start_timeout: float = 5.0,
        conversation_end_timeout: float = 1.0,
        conversation_max_duration: float = 15.0,
        **kwargs,
    ):
        """
        :param model: OpenAI model to use for audio transcription (default:
            ``whisper-1``).
        :param tts_plugin: Name of the TTS plugin to use for rendering the responses
            (default: ``tts.openai``).
        :param min_silence_secs: Minimum silence duration in seconds to detect
            the end of a conversation (default: 1.0 seconds).
        :param silence_threshold: Silence threshold in dBFS (default: -22).
            The value of 0 is the maximum amplitude, and -120 is associated to
            a silent or nearly silent audio, thus the higher the value, the more
            sensitive the silence detection will be (default: -22).
        :param sample_rate: Recording sample rate in Hz (default: 16000).
        :param frame_size: Recording frame size in samples (default: 16384).
            Note that it's important to make sure that ``frame_size`` /
            ``sample_rate`` isn't smaller than the minimum silence duration,
            otherwise the silence detection won't work properly.
        :param channels: Number of recording channels (default: 1).
        :param conversation_start_timeout: How long to wait for the
            conversation to start (i.e. the first non-silent audio frame to be
            detected) before giving up and stopping the recording (default: 5.0
            seconds).
        :param conversation_end_timeout: How many seconds of silence to wait
            after the last non-silent audio frame before stopping the recording
            (default: 1.5 seconds).
        :param conversation_max_duration: Maximum conversation duration in seconds
            (default: 15.0 seconds).
        """
        kwargs["tts_plugin"] = tts_plugin
        super().__init__(**kwargs)

        self._model = model
        self._min_silence_secs = min_silence_secs
        self._silence_threshold = silence_threshold
        self._sample_rate = sample_rate
        self._frame_size = frame_size
        self._channels = channels
        self._conversation_start_timeout = conversation_start_timeout
        self._conversation_end_timeout = conversation_end_timeout
        self._conversation_max_duration = conversation_max_duration
        self._start_recording_event = Event()
        self._disable_default_response = False
        self._recording_state = RecordingState(
            sample_rate=sample_rate,
            channels=channels,
            min_silence_secs=min_silence_secs,
            silence_threshold=silence_threshold,
        )

        self._recorder: Optional[AudioRecorder] = None

    def _to_audio_segment(self, data: np.ndarray) -> AudioSegment:
        return AudioSegment(
            data.tobytes(),
            frame_rate=self._sample_rate,
            sample_width=data.dtype.itemsize,
            channels=self._channels,
        )

    def _is_conversation_ended(self):
        # End if the recording has been stopped
        if not self._recorder or self._recorder.should_stop():
            return True

        # End if we reached the max conversation duration
        if self._recording_state.duration >= self._conversation_max_duration:
            return True

        # End if the conversation hasn't started yet and we reached the
        # conversation start timeout
        if (
            not self._recording_state.conversation_started
            and self._recording_state.duration >= self._conversation_start_timeout
        ):
            return True

        # End if the conversation has started and the user has been silent for
        # more than the conversation end timeout
        if (
            self._recording_state.conversation_started
            and self._recording_state.silence_duration >= self._conversation_end_timeout
        ):
            return True

        return False

    @property
    def _openai(self) -> OpenaiPlugin:
        openai: Optional[OpenaiPlugin] = get_plugin("openai")
        assert openai, (
            "OpenAI plugin not found. "
            "Please configure the `openai` plugin to use `assistant.openai`"
        )
        return openai

    def _get_prediction(self, audio: BytesIO) -> str:
        return self._openai.transcribe_raw(
            audio.getvalue(), extension='mp3', model=self._model
        )

    def _capture_audio(self, recorder: AudioRecorder):
        while not self.should_stop() and not self._is_conversation_ended():
            audio_data = recorder.read()
            if not audio_data:
                continue

            self._recording_state.add_audio(audio_data)

    def _audio_loop(self):
        while not self.should_stop():
            self._wait_recording_start()
            self._recording_state.reset()
            self._on_conversation_start()

            try:
                with AudioRecorder(
                    stop_event=self._should_stop,
                    sample_rate=self._sample_rate,
                    frame_size=self._frame_size,
                    channels=self._channels,
                ) as self._recorder:
                    self._capture_audio(self._recorder)
            finally:
                if self._recorder:
                    try:
                        self._recorder.stream.close()
                    except Exception as e:
                        self.logger.warning("Error closing the audio stream: %s", e)

                self._recorder = None

            if self._recording_state.is_silent():
                self._on_conversation_timeout()
            else:
                audio = self._recording_state.export_audio()
                text = self._get_prediction(audio)
                self._on_speech_recognized(text)

    def _wait_recording_start(self):
        self._start_recording_event.wait()
        self._start_recording_event.clear()

    def _start_conversation(self, *_, **__):
        self._disable_default_response = False
        self._recording_state.reset()
        self._start_recording_event.set()

    def _stop_conversation(self, *_, **__):
        self._disable_default_response = True
        super()._stop_conversation()
        self._recording_state.reset()
        if self._recorder:
            self._recorder.stop()

        self._on_conversation_end()

    def _on_speech_recognized(self, phrase: Optional[str]):
        super()._on_speech_recognized(phrase)

        # Dirty hack: wait a bit before stopping the conversation to make sure
        # that there aren't event hooks triggered in other threads that are
        # supposed to handle.
        if self.stop_conversation_on_speech_match:
            self.wait_stop(0.5)
            if self.should_stop():
                return

            if self._disable_default_response:
                self.logger.debug("Default response disabled, skipping response")
                return

        response = self._openai.get_response(phrase).output
        if response:
            self.render_response(response)
        else:
            self._on_no_response()

    @action
    def start_conversation(self, *_, **__):
        """
        Start a conversation with the assistant. The conversation will be
        automatically stopped after ``conversation_max_duration`` seconds of
        audio, or after ``conversation_start_timeout`` seconds of silence
        with no audio detected, or after ``conversation_end_timeout`` seconds
        after the last non-silent audio frame has been detected, or when the
        :meth:`.stop_conversation` method is called.
        """
        self._start_conversation()

    @action
    def mute(self, *_, **__):
        """
        .. note:: This plugin has no hotword detection, thus no continuous
            audio detection. Speech processing is done on-demand through the
            :meth:`.start_conversation` and :meth:`.stop_conversation` methods.
            Therefore, the :meth:`.mute` and :meth:`.unmute` methods are not
            implemented.
        """
        self.logger.warning(
            "assistant.openai.mute is not implemented because this plugin "
            "has no hotword detection, and the only way to stop a conversation "
            "is by calling stop_conversation()"
        )

    @action
    def unmute(self, *_, **__):
        """
        .. note:: This plugin has no hotword detection, thus no continuous
            audio detection. Speech processing is done on-demand through the
            :meth:`.start_conversation` and :meth:`.stop_conversation` methods.
            Therefore, the :meth:`.mute` and :meth:`.unmute` methods are not
            implemented.
        """
        self.logger.warning(
            "assistant.openai.unmute is not implemented because this plugin "
            "has no hotword detection, and the only way to start a conversation "
            "is by calling start_conversation()"
        )

    @action
    def send_text_query(self, text: str, *_, **__):
        """
        If the ``tts_plugin`` configuration is set, then the assistant will
        process the given text query through
        :meth:`platypush.plugins.openai.OpenaiPlugin.get_response` and render
        the response through the specified TTS plugin.

        :return: The response received from
            :meth:`platypush.plugins.openai.OpenaiPlugin.get_response`.
        """
        response = self._openai.get_response(text).output
        self.render_response(response)
        return response

    def main(self):
        while not self.should_stop():
            try:
                self._audio_loop()
            except Exception as e:
                self.logger.error("Audio loop error: %s", e, exc_info=True)
                self.wait_stop(5)
            finally:
                self.stop_conversation()

    def stop(self):
        self._stop_conversation()
        super().stop()



import logging

from platypush.config import Config

_logger = None


def logger():
    global _logger
    if not _logger:
        log_args = {
            'level': logging.INFO,
            'format': '%(asctime)-15s|%(levelname)5s|%(name)s|%(message)s',
        }

        level = (Config.get('backend.http') or {}).get('logging') or (
            Config.get('logging') or {}
        ).get('level')
        filename = (Config.get('backend.http') or {}).get('filename')

        if level:
            log_args['level'] = (
                getattr(logging, level.upper()) if isinstance(level, str) else level
            )
        if filename:
            log_args['filename'] = filename

        logging.basicConfig(**log_args)
        _logger = logging.getLogger('platypush:web')

    return _logger



from contextlib import contextmanager
import json
from typing import Generator, Optional

from tornado.web import stream_request_body

from platypush.backend.http.app.utils import send_request
from platypush.config import Config

from .. import StreamingRoute


@stream_request_body
class SoundRoute(StreamingRoute):
    """
    Route for audio streams.
    """

    _redis_queue_prefix = f'_platypush/{Config.get("device_id") or ""}/sound'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._audio_headers_written: bool = False
        """Send the audio file headers before we send the first audio frame."""

    @classmethod
    def path(cls) -> str:
        return r"/sound/stream\.?([a-zA-Z0-9_]+)?"

    @contextmanager
    def _audio_stream(self, **kwargs) -> Generator[None, None, None]:
        send_request(
            'sound.record',
            dtype='int16',
            **kwargs,
        )

        yield
        send_request('sound.stop_recording')

    @classmethod
    def _get_redis_queue(cls, *_, device: Optional[str] = None, **__) -> str:
        return '/'.join([cls._redis_queue_prefix, *([device] if device else [])])

    def _get_args(self, **kwargs):
        kwargs.update({k: v[0].decode() for k, v in self.request.arguments.items()})
        device = kwargs.get('device')
        return {
            'device': device,
            'sample_rate': int(kwargs.get('sample_rate', 44100)),
            'blocksize': int(kwargs.get('blocksize', 512)),
            'latency': float(kwargs.get('latency', 0)),
            'channels': int(kwargs.get('channels', 1)),
            'format': kwargs.get('format', 'wav'),
            'redis_queue': kwargs.get('redis_queue', self._get_redis_queue(device)),
        }

    @staticmethod
    def _content_type_by_extension(extension: str) -> str:
        if extension == 'mp3':
            return 'audio/mpeg'
        if extension == 'ogg':
            return 'audio/ogg'
        if extension == 'wav':
            return 'audio/wav;codec=pcm'
        if extension == 'flac':
            return 'audio/flac'
        if extension == 'aac':
            return 'audio/aac'
        return 'application/octet-stream'

    def get(self, extension: Optional[str] = None) -> None:
        ext = extension or 'wav'
        args = self._get_args(format=ext)

        try:
            with self._audio_stream(**args):
                self.set_header('Content-Type', self._content_type_by_extension(ext))
                self.forward_stream(**args)

            self.finish()
        except AssertionError as e:
            self.set_header("Content-Type", "application/json")
            self.set_status(400, str(e))
            self.finish(json.dumps({"error": str(e)}))
        except Exception as e:
            self.set_header("Content-Type", "application/json")
            self.logger.exception(e)
            self.set_status(500, str(e))
            self.finish(json.dumps({"error": str(e)}))



from abc import ABC, abstractmethod
import asyncio
import concurrent.futures
from typing import Callable, Coroutine, Optional

from .._types import Errors
from ._base import XmppBaseMixin


# pylint: disable=too-few-public-methods
class XmppAsyncMixin(XmppBaseMixin, ABC):
    """
    This mixin provides a common interface for aioxmpp's asyncio interface.
    """

    @abstractmethod
    def __init__(
        self, *args, loop: Optional[asyncio.AbstractEventLoop] = None, **kwargs
    ):
        self._loop = loop
        super().__init__(*args, **kwargs)

    def _async_run(
        self,
        coro: Callable[..., Coroutine],
        *args,
        timeout: Optional[float] = XmppBaseMixin.DEFAULT_TIMEOUT,
        wait_result: bool = True,
        **kwargs,
    ):
        """
        Utility method to call an async action from the thread of the parent
        action.
        """
        assert self._loop, Errors.LOOP
        fut = asyncio.run_coroutine_threadsafe(coro(*args, **kwargs), self._loop)

        if wait_result:
            err = None

            try:
                return fut.result(timeout)
            except (TimeoutError, concurrent.futures.TimeoutError) as e:
                self.logger.warning(
                    'Call to %s timed out after %f seconds', coro, timeout
                )
                err = e
            except Exception as e:
                self.logger.warning('Call to %s failed: %s', coro, e)
                self.logger.exception(e)
                err = e
            finally:
                assert not err, str(err)

        return None



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional

from platypush.schemas.dataclasses import percent_field


@dataclass
class Process:
    """
    System process data class.
    """

    pid: int = field(
        metadata={
            'metadata': {
                'description': 'Process PID',
                'example': 12345,
            }
        }
    )

    name: str = field(
        metadata={
            'metadata': {
                'description': 'Process name',
                'example': 'python',
            }
        }
    )

    parent_pid: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'PID of the parent process',
                'example': 1000,
            }
        }
    )

    username: str = field(
        metadata={
            'metadata': {
                'description': 'Username that owns the process',
                'example': 'root',
            }
        }
    )

    command_line: List[str] = field(
        metadata={
            'metadata': {
                'description': 'Command line of the process',
                'example': ['/usr/bin/python', '-m', 'platypush'],
            }
        }
    )

    status: str = field(
        metadata={
            'metadata': {
                'description': 'Process status',
                'example': 'running',
            }
        }
    )

    terminal: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Terminal the process is running on',
                'example': 'pts/1',
            }
        }
    )

    current_directory: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Current directory of the process',
                'example': '/root',
            }
        }
    )

    started: Optional[datetime] = field(
        metadata={
            'metadata': {
                'description': 'When the process started',
            }
        }
    )

    cpu_percent: float = percent_field(
        metadata={
            'metadata': {
                'description': 'Percentage of CPU used by the process, between 0 and 1',
                'example': 0.1,
            }
        }
    )

    memory_percent: float = percent_field(
        metadata={
            'metadata': {
                'description': 'Percentage of memory used by the process, between 0 and 1',
                'example': 0.05,
            }
        }
    )



from enum import Enum
from dataclasses import dataclass
from typing import Optional


class BaseBluetoothClass(Enum):
    """
    Base enum to model Bluetooth device/service classes.
    """

    def __repr__(self) -> str:
        """
        :return: The enum class formatted as ``<name: <value>``.
        """
        return f'<{self.__class__.__name__}.{self.name}: {str(self)}>'

    def __str__(self) -> str:
        """
        :return Only the readable string value of the class.
        """
        return self.value.name


@dataclass
class ClassProperty:
    """
    Models a Bluetooth class property.

    Given a Bluetooth class as a 24-bit unsigned integer, this class models the
    filter that should be applied to the class to tell if the device exposes the
    property.
    """

    name: str
    """ The name of the property. """
    bitmask: int
    """ Bitmask used to select the property bits from the class. """
    bit_shift: int = 0
    """ Number of bits to shift the class value after applying the bitmask. """
    match_value: int = 1
    """
    This should be the result of the bitwise filter for the property to match.
    """
    parent: Optional[BaseBluetoothClass] = None
    """
    Parent class property, if this is a minor property. If this is the case,
    then the advertised class value should also match the parent property.
    """

    def matches(self, cls: int) -> bool:
        """
        Match function.

        :param cls: The Bluetooth composite class value.
        :return: True if the class matches the property.
        """
        if self.parent and not self.parent.value.matches(cls):
            return False
        return ((cls & self.bitmask) >> self.bit_shift) == self.match_value



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Type

from platypush.common.sensors import Numeric
from platypush.entities.acceleration import Accelerometer
from platypush.entities.devices import Device
from platypush.entities.distance import DistanceSensor
from platypush.entities.illuminance import IlluminanceSensor
from platypush.entities.magnetism import Magnetometer
from platypush.entities.pressure import PressureSensor
from platypush.entities.sensors import BinarySensor, RawSensor, Sensor
from platypush.entities.temperature import TemperatureSensor
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


@dataclass
class SensorEntityMapping:
    """
    Maps sensor attributes returned by
    :meth:`SensorEnvirophatPlugin.get_measurement` to native sensor entities.
    """

    name: str
    entity_type: Type[Sensor]
    unit: Optional[str] = None
    kwargs: Dict[str, Any] = field(default_factory=dict)

    @property
    def id(self):
        """
        Standard format for an Envirophat sensor external ID.
        """
        return f'envirophat:{self.name}'


_sensor_entity_mappings: Dict[str, SensorEntityMapping] = {
    mapping.name: mapping
    for mapping in [
        SensorEntityMapping(
            name='temperature',
            entity_type=TemperatureSensor,
            unit='°C',
        ),
        SensorEntityMapping(
            name='pressure',
            entity_type=PressureSensor,
            unit='Pa',
        ),
        SensorEntityMapping(
            name='altitude',
            entity_type=DistanceSensor,
            unit='m',
        ),
        SensorEntityMapping(
            name='illuminance',
            entity_type=IlluminanceSensor,
            unit='lux',
        ),
        SensorEntityMapping(
            name='accelerometer',
            entity_type=Accelerometer,
        ),
        SensorEntityMapping(
            name='magnetometer',
            entity_type=Magnetometer,
        ),
        SensorEntityMapping(
            name='analog',
            entity_type=RawSensor,
        ),
        SensorEntityMapping(
            name='leds',
            entity_type=BinarySensor,
        ),
    ]
}


# pylint: disable=too-many-ancestors
class SensorEnvirophatPlugin(SensorPlugin):
    """
    Plugin to interact with a `Pimoroni enviropHAT
    <https://shop.pimoroni.com/products/enviro-phat>`_ device.

    You can use an enviropHAT device to read e.g. temperature, pressure,
    altitude, accelerometer, magnetometer and luminosity data.
    """

    @action
    def get_measurement(self, *_, qnh: float = 1020.0, **__):
        """
        :param: qnh: Local value for atmospheric pressure adjusted to sea level
            (default: 1020)

        :returns: dict. Example:

        .. code-block:: python

           output = {
               "temperature": 21.0,   # Celsius
               "pressure": 101555.08, # pascals
               "altitude": 10,        # meters
               "luminosity": 426,     # lumens

               # Measurements from the custom analog channels
               "analog": [0.513, 0.519, 0.531, 0.528],

               "accelerometer": {
                   "x": -0.000915,
                   "y": 0.0760,
                   "z": 1.026733
               },
               "magnetometer": {
                   "x": -2297,
                   "y": 1226,
                   "z": -7023
               },
           }

        """

        import envirophat

        ret = {}
        weather = envirophat.weather
        light = envirophat.light
        accelerometer = envirophat.motion.accelerometer()
        magnetometer = envirophat.motion.magnetometer()
        leds = envirophat.leds
        analog = envirophat.analog

        weather.update()

        ret['temperature'] = weather.temperature()
        ret['pressure'] = weather.pressure()
        ret['altitude'] = weather.altitude(qnh=qnh)
        ret['luminosity'] = light.light()
        ret['accelerometer'] = {v: getattr(accelerometer, v) for v in ['x', 'y', 'z']}
        ret['magnetometer'] = {v: getattr(magnetometer, v) for v in ['x', 'y', 'z']}
        ret['analog'] = list(analog.read_all())
        ret['leds'] = leds.is_on()

        return ret

    def transform_entities(self, entities: Dict[str, Numeric]) -> List[Device]:
        return [
            Device(
                id='envirophat',
                name='EnviroPHAT',
                children=[
                    _sensor_entity_mappings[sensor].entity_type(
                        id=_sensor_entity_mappings[sensor].id,
                        name=_sensor_entity_mappings[sensor].name,
                        unit=_sensor_entity_mappings[sensor].unit,
                        **_sensor_entity_mappings[sensor].kwargs,
                    )
                    for sensor, value in entities.items()
                    if value is not None
                ],
            )
        ]


# vim:sw=4:ts=4:et:



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional


@dataclass
class User:
    """
    System user wrapper.
    """

    username: str = field(
        metadata={
            'metadata': {
                'description': 'Username',
                'example': 'root',
            }
        }
    )

    terminal: Optional[str] = field(
        metadata={
            'metadata': {
                'description': 'Identifier of the terminal the user is connected to',
                'example': 'pts/1',
            }
        }
    )

    started: Optional[datetime] = field(
        metadata={
            'metadata': {
                'description': 'When the user session started',
            }
        }
    )

    pid: Optional[int] = field(
        metadata={
            'metadata': {
                'description': 'PID of the process that holds the session',
                'example': 12345,
            }
        }
    )
